import numpy as np
import scipy.constants as spc
from openmdao.api import ExplicitComponent
from scipy.optimize import minimize

import wisdem.commonse.utilities as util
import wisdem.pyframe3dd.pyframe3dd as pyframe3dd
from wisdem.commonse.constants import gravity


# This isn't used, but keeping around the code for now
def enforce_length(x, y, z, L0):
    r0 = np.sum(L0)
    xn = x.copy()
    yn = y.copy()
    zn = z.copy()
    rn = util.arc_length(np.c_[xn, yn, zn])

    counter = 0
    while np.abs(rn[-1] - r0) > 1e-2 and counter < 100:
        counter += 1
        L = np.diff(rn)
        xn[1:] = (L0 / L) * (xn[1:] - xn[:-1]) + xn[:-1]
        zn[1:] = (L0 / L) * (zn[1:] - zn[:-1]) + zn[:-1]
        rn = util.arc_length(np.c_[xn, yn, zn])
    return xn, zn


class RailTransport(ExplicitComponent):
    # Openmdao component to simulate a rail transport of a wind turbine blade
    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        rotorse_options = self.options["modeling_options"]["WISDEM"]["RotorSE"]
        self.n_span = n_span = rotorse_options["n_span"]
        self.n_xy = n_xy = rotorse_options["n_xy"]  # Number of coordinate points to describe the airfoil geometry

        # Rail configuration
        self.add_input(
            "horizontal_angle_deg",
            val=13.0,
            units="deg",
            desc="Angle of horizontal turn (defined for an chord of 100 feet)",
        )
        self.add_input(
            "min_vertical_radius",
            val=609.6,
            units="m",
            desc="Minimum radius of a vertical curvature (hill or sag) (2000 feet)",
        )
        # self.add_input("lateral_clearance", val=6.7056, units="m", desc="Clearance profile horizontal (22 feet)")
        # self.add_input("vertical_clearance", val=7.0104, units="m", desc="Clearance profile vertical (23 feet)")
        self.add_input("lateral_clearance", val=5.4864, units="m", desc="Clearance profile horizontal (18 feet)")
        self.add_input("vertical_clearance", val=6.096, units="m", desc="Clearance profile vertical (20 feet)")
        self.add_input(
            "deck_height", val=1.19, units="m", desc="Height of the deck of the flatcar from the rails (4 feet)"
        )
        self.add_input("max_strains", val=3500.0 * 1.0e-6, desc="Max allowable strains during transport")
        self.add_input("max_LV", val=0.5, desc="Max allowable ratio between lateral and vertical forces")
        self.add_input(
            "max_flatcar_weight_4axle", val=129727.31, units="kg", desc="Max mass of an 4-axle flatcar (286000 lbm)"
        )
        self.add_input(
            "max_flatcar_weight_8axle", val=217724.16, units="kg", desc="Max mass of an 8-axle flatcar (480000 lbm)"
        )
        self.add_input("max_root_rot_deg", val=15.0, units="deg", desc="Max degree of angle at blade root")
        self.add_input("flatcar_tc_length", val=20.12, units="m", desc="Flatcar truck center to truck center length")

        # Input - Outer blade geometry
        self.add_input(
            "blade_ref_axis",
            val=np.zeros((n_span, 3)),
            units="m",
            desc="2D array of the coordinates (x,y,z) of the blade reference axis, defined along blade span. The coordinate system is the one of BeamDyn: it is placed at blade root with x pointing the suction side of the blade, y pointing the trailing edge and z along the blade span. A standard configuration will have negative x values (prebend), if swept positive y values, and positive z values.",
        )
        self.add_input(
            "theta",
            val=np.zeros(n_span),
            units="deg",
            desc="Twist angle at each section (positive decreases angle of attack)",
        )
        self.add_input("chord", val=np.zeros(n_span), units="m", desc="chord length at each section")
        self.add_input(
            "pitch_axis",
            val=np.zeros(n_span),
            desc="1D array of the chordwise position of the pitch axis (0-LE, 1-TE), defined along blade span.",
        )
        self.add_input(
            "coord_xy_interp",
            val=np.zeros((n_span, n_xy, 2)),
            desc="3D array of the non-dimensional x and y airfoil coordinates of the airfoils interpolated along span for n_span stations. The leading edge is place at x=0 and y=0.",
        )

        # Inputs - Distributed beam properties
        self.add_input("A", val=np.zeros(n_span), units="m**2", desc="airfoil cross section material area")
        self.add_input("EA", val=np.zeros(n_span), units="N", desc="axial stiffness")
        self.add_input(
            "EIxx",
            val=np.zeros(n_span),
            units="N*m**2",
            desc="edgewise stiffness (bending about :ref:`x-axis of airfoil aligned coordinate system <blade_airfoil_coord>`)",
        )
        self.add_input(
            "EIyy",
            val=np.zeros(n_span),
            units="N*m**2",
            desc="flatwise stiffness (bending about y-axis of airfoil aligned coordinate system)",
        )
        self.add_input("EIxy", val=np.zeros(n_span), units="N*m**2", desc="coupled flap-edge stiffness")
        self.add_input(
            "GJ",
            val=np.zeros(n_span),
            units="N*m**2",
            desc="torsional stiffness (about axial z-axis of airfoil aligned coordinate system)",
        )
        self.add_input("rhoA", val=np.zeros(n_span), units="kg/m", desc="mass per unit length")
        self.add_input("rhoJ", val=np.zeros(n_span), units="kg*m", desc="polar mass moment of inertia per unit length")
        self.add_input(
            "x_sc",
            val=np.zeros(n_span),
            units="m",
            desc="X-coordinate of the shear-center offset with respect to the XR-YR axes",
        )
        self.add_input(
            "y_sc",
            val=np.zeros(n_span),
            units="m",
            desc="Chordwise offset of the section shear-center with respect to the reference frame, XR-YR",
        )

        # Outputs
        self.add_output(
            "constr_LV_4axle_horiz",
            val=np.zeros(2),
            desc="Constraint for max L/V for a 4-axle flatcar on horiz curves, violated when bigger than 1",
        )
        self.add_output(
            "constr_LV_8axle_horiz",
            val=np.zeros(2),
            desc="Constraint for max L/V for an 8-axle flatcar on horiz curves, violated when bigger than 1",
        )
        self.add_output(
            "constr_LV_4axle_vert",
            val=np.zeros(2),
            desc="Constraint for max L/V for a 4-axle flatcar on vert curves, violated when bigger than 1",
        )
        self.add_output(
            "constr_LV_8axle_vert",
            val=np.zeros(2),
            desc="Constraint for max L/V for an 8-axle flatcar on vert curves, violated when bigger than 1",
        )
        self.add_output(
            "constr_strainPS", val=np.zeros(n_span), desc="Strain along pressure side of blade on a horizontal curve"
        )
        self.add_output(
            "constr_strainSS", val=np.zeros(n_span), desc="Strain along suction side of blade on a horizontal curve"
        )
        self.add_output(
            "constr_strainLE",
            val=np.zeros((n_span, 2)),
            desc="Strain along leading edge side of blade on a vertical curve",
        )
        self.add_output(
            "constr_strainTE",
            val=np.zeros((n_span, 2)),
            desc="Strain along trailing edge side of blade on a vertical curve",
        )

    def compute(self, inputs, outputs):
        PBEAM = False
        _8axle = False

        # Unpack inputs
        x_ref = inputs["blade_ref_axis"][:, 0]  # from PS to SS
        y_ref = inputs["blade_ref_axis"][:, 1]  # from LE to TE
        z_ref = inputs["blade_ref_axis"][:, 2]  # from root to tip
        r = util.arc_length(inputs["blade_ref_axis"])
        blade_length = r[-1]
        theta = inputs["theta"]
        chord = inputs["chord"]
        x_sc = inputs["x_sc"]
        y_sc = inputs["y_sc"]
        A = inputs["A"]
        rhoA = inputs["rhoA"]
        rhoJ = inputs["rhoJ"]
        GJ = inputs["GJ"]
        EA = inputs["EA"]
        EIxx = inputs["EIxx"]  # edge (rotation about x)
        EIyy = inputs["EIyy"]  # flap (rotation about y)
        EIxy = inputs["EIxy"]
        lateral_clearance = 0.5 * float(inputs["lateral_clearance"][0])
        vertical_clearance = float(inputs["vertical_clearance"][0])
        max_strains = float(inputs["max_strains"][0])
        max_rot = float(inputs["max_root_rot_deg"][0])
        max_LV = float(inputs["max_LV"][0])
        if _8axle:
            mass_car = float(inputs["max_flatcar_weight_8axle"][0])
        else:
            mass_car = float(inputs["max_flatcar_weight_4axle"][0])
        flatcar_tc_length = float(inputs["flatcar_tc_length"][0])

        # ------- Get turn radius geometry for horizontal and vertical curves
        # Horizontal turns- defined as a degree of arc assuming a 100ft "chord"
        # https://trn.trains.com/railroads/ask-trains/2011/01/measuring-track-curvature
        angleH_rad = np.deg2rad(inputs["horizontal_angle_deg"][0])
        r_curveH = spc.foot * 100.0 / (2.0 * np.sin(0.5 * angleH_rad))
        arcsH = r / r_curveH

        # Vertical curves on hills and sags defined directly by radius
        r_curveV = inputs["min_vertical_radius"][0]
        # ----------

        # ---------- Put airfoil cross sections into principle axes
        # Determine principal C.S. (with swap of x, y for profile c.s.)
        EIxx_cs, EIyy_cs = EIyy.copy(), EIxx.copy()
        x_sc_cs, y_sc_cs = y_sc.copy(), x_sc.copy()
        EIxy_cs = EIxy.copy()

        # translate to elastic center
        EIxx_cs -= y_sc_cs**2 * EA
        EIyy_cs -= x_sc_cs**2 * EA
        EIxy_cs -= x_sc_cs * y_sc_cs * EA

        # get rotation angle
        alpha = 0.5 * np.arctan2(2 * EIxy_cs, EIyy_cs - EIxx_cs)

        # get moments and positions in principal axes
        EI11 = EIxx_cs - EIxy_cs * np.tan(alpha)
        EI22 = EIyy_cs + EIxy_cs * np.tan(alpha)
        ca = np.cos(alpha)
        sa = np.sin(alpha)

        def rotate(x, y):
            x2 = x * ca + y * sa
            y2 = -x * sa + y * ca
            return x2, y2

        # Now store alpha for later use in degrees
        alpha = np.rad2deg(alpha)
        # -------------------

        # ---------- Frame3dd blade prep
        # Nodes: Prep data, but node x,y,z will shift for vertical and horizontal curves
        rad = np.zeros(self.n_span)  # 'radius' of rigidity at node- set to zero
        inode = 1 + np.arange(self.n_span)  # Node numbers (will convert to 1-based indexing later)
        L = np.diff(r)

        # Reactions: attachment at root
        ireact = np.array([inode[0]])
        rigid = 1e16
        pinned = rigid * np.ones(1)
        reactions = pyframe3dd.ReactionData(ireact, pinned, pinned, pinned, pinned, pinned, pinned, float(rigid))

        # Element data
        ielem = np.arange(1, self.n_span)  # Element Numbers
        N1 = np.arange(1, self.n_span)  # Node number start
        N2 = np.arange(2, self.n_span + 1)  # Node number finish
        E = EA / A
        rho = rhoA / A
        J = rhoJ / rho
        G = GJ / J
        Ix = EIyy / E if PBEAM else EI11 / E
        Iy = EIxx / E if PBEAM else EI22 / E
        Asx = Asy = 1e-6 * np.ones(ielem.shape)  # Unused when shear=False

        # Have to convert nodal values to find average at center of element
        Abar, _ = util.nodal2sectional(A)
        Ebar, _ = util.nodal2sectional(E)
        rhobar, _ = util.nodal2sectional(rho)
        Jbar, _ = util.nodal2sectional(J)
        Gbar, _ = util.nodal2sectional(G)
        Ixbar, _ = util.nodal2sectional(Ix)
        Iybar, _ = util.nodal2sectional(Iy)

        # Angle of element principal axes relative to global coordinate system
        # Global c.s. is blade with z from root to tip, y from ss to ps, and x from LE to TE (TE points up)
        # Local element c.s. is airfoil (twist + principle rotation)
        if PBEAM:
            roll = np.zeros(theta.shape)
        else:
            roll, _ = util.nodal2sectional(theta + alpha)

        elements = pyframe3dd.ElementData(ielem, N1, N2, Abar, Asx, Asy, Jbar, Ixbar, Iybar, Ebar, Gbar, roll, rhobar)

        # Frame3dd options: no need for shear, axial stiffening, or higher resolution force calculations
        options = pyframe3dd.Options(False, False, -1)
        # -----------

        # ------ Airfoil positions at which to measure strain
        # Find the cross sectional points furthest from the elastic center at each spanwise location to be used for strain measurement
        xps = np.zeros(self.n_span)
        xss = np.zeros(self.n_span)
        yps = np.zeros(self.n_span)
        yss = np.zeros(self.n_span)
        xle = np.zeros(self.n_span)
        xte = np.zeros(self.n_span)
        yle = np.zeros(self.n_span)
        yte = np.zeros(self.n_span)

        for i in range(self.n_span):
            ## Rotate the profiles to the blade reference system
            profile_i = inputs["coord_xy_interp"][i, :, :]
            profile_i_rot = np.column_stack(
                util.rotate(inputs["pitch_axis"][i], 0.0, profile_i[:, 0], profile_i[:, 1], np.radians(theta[i]))
            )
            # normalize
            profile_i_rot[:, 0] -= min(profile_i_rot[:, 0])
            profile_i_rot = profile_i_rot / max(profile_i_rot[:, 0])
            profile_i_rot_precomp = profile_i_rot.copy()
            idx_s = 0
            idx_le_precomp = np.argmax(profile_i_rot_precomp[:, 0])
            if idx_le_precomp != 0:
                if profile_i_rot_precomp[0, 0] == profile_i_rot_precomp[-1, 0]:
                    idx_s = 1
                profile_i_rot_precomp = np.row_stack(
                    (profile_i_rot_precomp[idx_le_precomp:], profile_i_rot_precomp[idx_s:idx_le_precomp, :])
                )
            profile_i_rot_precomp[:, 1] -= profile_i_rot_precomp[np.argmin(profile_i_rot_precomp[:, 0]), 1]

            # # renormalize
            profile_i_rot_precomp[:, 0] -= min(profile_i_rot_precomp[:, 0])
            profile_i_rot_precomp = profile_i_rot_precomp / max(profile_i_rot_precomp[:, 0])

            if profile_i_rot_precomp[-1, 0] != 1.0:
                profile_i_rot_precomp = np.row_stack((profile_i_rot_precomp, profile_i_rot_precomp[0, :]))

            xnode = profile_i_rot_precomp[:, 0]
            xnode_pa = xnode - inputs["pitch_axis"][i]
            ynode = profile_i_rot_precomp[:, 1]
            theta_rad = theta[i] * np.pi / 180.0

            xnode_no_theta = xnode_pa * np.cos(-theta_rad) - ynode * np.sin(-theta_rad)
            ynode_no_theta = xnode_pa * np.sin(-theta_rad) + ynode * np.cos(-theta_rad)

            xnode_dim_no_theta = xnode_no_theta * chord[i]
            ynode_dim_no_theta = ynode_no_theta * chord[i]

            xnode_dim = xnode_dim_no_theta * np.cos(theta_rad) - ynode_dim_no_theta * np.sin(theta_rad)
            ynode_dim = xnode_dim_no_theta * np.sin(theta_rad) + ynode_dim_no_theta * np.cos(theta_rad)

            yss[i] = max(ynode_dim) - y_sc_cs[i]
            yps[i] = y_sc_cs[i] - min(ynode_dim)
            xte[i] = max(xnode_dim) - x_sc_cs[i]
            xle[i] = x_sc_cs[i] - min(xnode_dim)

        # Put these sectional points in airfoil principle directions
        xps_cs, yps_cs = yps, xps
        xss_cs, yss_cs = yss, xss

        ps1, ps2 = rotate(xps_cs, yps_cs)
        ss1, ss2 = rotate(xss_cs, yss_cs)

        # ----------------

        # -------- Horizontal curve where we select blade support nodes on flat cars
        # Gravity field orientation
        gy = -gravity
        gx = gz = 0.0

        # Set clearance boundary
        r_envelopeH = r_curveH + lateral_clearance * np.array([-1, 1])

        # Use blade shape and clearance envelope to determine node position limits
        r_envelopeH_inner1 = r_envelopeH.min() + yss
        r_envelopeH_inner2 = r_envelopeH.min() + yps

        r_envelopeH_outer1 = r_envelopeH.max() - yps
        r_envelopeH_outer2 = r_envelopeH.max() - yss

        # Find rotation angles that keep blade within inner boundary
        # Function that does the structural analysis to be called during optimization
        def rotate_blade(ang):
            # Node location starting points when curving towards SS
            # (towards the LEFT with LE pointed down and standing at the root looking at tip)
            x_rot1, z_rot1 = util.rotate(r_curveH, 0.0, r_curveH + x_ref, z_ref, ang[0])

            # Node location starting points when curving towards PS
            # (towards the RIGHT with LE pointed down and standing at the root looking at tip)
            x_rot2, z_rot2 = util.rotate(-r_curveH, 0.0, -r_curveH + x_ref, z_ref, ang[1])

            # Check solved blade shape against envelope
            r_check1 = np.sqrt(x_rot1**2 + z_rot1**2)
            r_check2 = np.sqrt(x_rot2**2 + z_rot2**2)

            # Formulate as constraints for SLSQP
            cboundary = np.sum(np.maximum(r_envelopeH_inner1 - r_check1, 0.0)) + np.sum(
                np.maximum(r_envelopeH_inner2 - r_check2, 0.0)
            )

            return -cboundary

        # Initiliaze scipy minimization to find the initial blade root flapwise angle - biggest deflection that keeps blade away from inner envelope
        const = {}
        const["type"] = "ineq"
        const["fun"] = rotate_blade

        bounds = [np.deg2rad(max_rot) * np.r_[0, 1], np.deg2rad(max_rot) * np.r_[-1, 0]]
        x0 = np.deg2rad(max_rot * np.array([1.0, -1.0]))
        result = minimize(lambda x: -np.sum(np.abs(x)), x0, method="slsqp", bounds=bounds, tol=1e-3, constraints=const)

        # Now rotate blade at optimized angle
        # x_rot1, z_rot1 = util.rotate(r_curveH, 0.0, r_curveH + x_ref, z_ref, result.x[0])
        # nodes1 = pyframe3dd.NodeData(inode, x_rot1, y_ref, z_rot1, rad)

        # Node location starting points when curving towards PS
        # (towards the RIGHT with LE pointed down and standing at the root looking at tip)
        # x_rot2, z_rot2 = util.rotate(-r_curveH, 0.0, -r_curveH + x_ref, z_ref, result.x[1])
        # print('Angle: ', np.rad2deg(result.x[1]))
        # x_rot2, z_rot2 = util.rotate(-r_curveH, 0.0, -r_curveH + x_ref, z_ref, np.deg2rad(-20))
        # nodes2 = pyframe3dd.NodeData(inode, x_rot2, y_ref, z_rot2, rad)

        x_rail_inner = np.linspace(0.0, 2.0 * r_envelopeH.min(), 10000)
        y_rail_inner = np.sqrt(r_envelopeH.min() ** 2.0 - (x_rail_inner - r_envelopeH.min()) ** 2.0)
        x_rail_outer = np.linspace(0.0, 2.0 * r_envelopeH.max(), 10000)
        y_rail_outer = np.sqrt(r_envelopeH.max() ** 2.0 - (x_rail_outer - r_envelopeH.max()) ** 2.0)

        # # # # Undeflected transport
        # import matplotlib.pyplot as plt
        # fig, ax = plt.subplots(1, 1, figsize=(3.5,8))
        # ax.plot(x_rail_inner+2*lateral_clearance, y_rail_inner, "k:", label='Lateral clearance')
        # ax.plot(x_rail_outer, y_rail_outer, "k:")
        # # ax.plot(x_rot1+x_rot1[0]+lateral_clearance, z_rot1, label='rot1')
        # ax.plot(x_rot2-x_rot2[0]+lateral_clearance + yss, z_rot2, color="tab:red", label='Blade')
        # ax.plot(x_rot2-x_rot2[0]+lateral_clearance - yps, z_rot2, color="tab:red")
        # ax.set_xlim([0,50])
        # ax.set_ylim([-0,110])
        # ax.set_xlabel('m', weight='bold')
        # ax.set_ylabel('m', weight='bold')
        # # ax.axis('equal')
        # plt.grid(color=[0.8,0.8,0.8], linestyle='--')
        # plt.subplots_adjust(bottom = 0.15, left = 0.2)
        # ax.legend()
        # plt.savefig('/Users/pbortolo/work/9_Presentations/FY22/BAR/RailReview/23deg_violate.pdf')
        # # plt.savefig('/Users/pbortolo/work/9_Presentations/FY22/BAR/RailReview/23deg.pdf')
        # plt.show()
        # exit()

        def get_max_force_h(FrIn):
            # Objective function to minimize the reaction force of the first flatcat, which holds blade root, during a lateral curve
            Fr = FrIn[2:].reshape((self.n_span - 1, 2))
            RF_flatcar_1 = np.zeros(2)
            for k in range(2):
                q_iter = np.r_[0.0, Fr[:, k]]
                V_iter = np.zeros(self.n_span)
                M_iter = np.zeros(self.n_span)
                for i in range(self.n_span):
                    V_iter[i] = np.trapz(q_iter[i:], r[i:])
                for i in range(self.n_span):
                    M_iter[i] = np.trapz(V_iter[i:], r[i:])

                RF_flatcar_1[k] = 0.5 * V_iter[0] + M_iter[0] / flatcar_tc_length

            return np.sum(abs(RF_flatcar_1) * 1.0e-5)

        # Function that does the structural analysis to be called during optimization
        def run_hcurve(FrIn, optFlag=True, final=False):
            angle = FrIn[0:2]
            if final:
                print("Flapwise rotations at blade root: ", angle)

            # Node location starting points when curving towards SS
            # (towards the LEFT with LE pointed down and standing at the root looking at tip)
            x_rot1, z_rot1 = util.rotate(r_curveH, 0.0, r_curveH + x_ref, z_ref, angle[0])
            nodes1 = pyframe3dd.NodeData(inode, x_rot1, y_ref, z_rot1, rad)

            # Node location starting points when curving towards PS
            # (towards the RIGHT with LE pointed down and standing at the root looking at tip)
            x_rot2, z_rot2 = util.rotate(-r_curveH, 0.0, -r_curveH + x_ref, z_ref, angle[1])
            nodes2 = pyframe3dd.NodeData(inode, x_rot2, y_ref, z_rot2, rad)

            # Initialize Frame3dd objects
            blade1 = pyframe3dd.Frame(nodes1, reactions, elements, options)
            blade2 = pyframe3dd.Frame(nodes2, reactions, elements, options)

            Fr = FrIn[2:].reshape((self.n_span - 1, 2))

            # Will only worry about radial loads
            Fy = Mx = My = Mz = np.zeros(self.n_span - 1)

            # Output containers
            RF_derailH = np.zeros(2)  # Derailment reaction force
            r_check_outer = np.zeros((self.n_span, 2))  # Envelope constraint
            r_check_inner = np.zeros((self.n_span, 2))  # Envelope constraint
            strainPS = np.zeros((self.n_span, 2))
            strainSS = np.zeros((self.n_span, 2))

            # Look over bend to PS/SS cases
            for k in range(2):
                if k == 0:
                    blade, r_outer, r_inner, angs = blade1, r_envelopeH_outer1, r_envelopeH_inner1, arcsH[1:]
                else:
                    blade, r_outer, r_inner, angs = blade2, r_envelopeH_outer2, r_envelopeH_inner2, np.pi - arcsH[1:]

                # Load case: gravity + blade bending to conform to outer boundary
                load = pyframe3dd.StaticLoadCase(gx, gy, gz)

                # Put radial loads in x,z plane
                Fx = -1e4 * Fr[:, k] * np.cos(angs)
                Fz = -1e4 * Fr[:, k] * np.sin(angs)
                load.changePointLoads(inode[1:], Fx, Fy, Fz, Mx, My, Mz)

                # Add load case to Frame3DD objects and run
                blade.clearLoadCases()
                blade.addLoadCase(load)

                # Run the case
                displacements, forces, forces_rxn, _, _, _ = blade.run()
                L0 = np.diff(util.arc_length(np.vstack((blade.nx, blade.nz)).T))
                x = blade.nx + displacements.dx[0, :]
                y = blade.ny + displacements.dy[0, :]
                z = blade.nz + displacements.dz[0, :]
                x_defl, z_defl = enforce_length(x, y, z, L0)

                # Check solved blade shape against envelope

                r_check_outer[:, k] = np.sqrt(x_defl**2 + z_defl**2) - r_outer
                r_check_inner[:, k] = r_inner - np.sqrt(x_defl**2 + z_defl**2)

                # if final and k==1:
                #     print("Undeflected blade length: ", util.arc_length(np.vstack((blade.nx, blade.nz)).T)[-1])
                #     print("Corrected blade length: ", util.arc_length(np.vstack((x_defl, z_defl)).T)[-1])
                #     # Deflected transport
                #     fig, ax = plt.subplots(1, 1, figsize=(3.5,8))
                #     ax.plot(x_rail_inner+2*lateral_clearance, y_rail_inner, "k:", label='Lateral clearance')
                #     ax.plot(x_rail_outer, y_rail_outer, "k:")
                #     # ax.plot(x_rot1+x_rot1[0]+lateral_clearance, z_rot1, label='rot1')
                #     ax.plot(x_defl + yss - blade.nx[0] + lateral_clearance, z_defl, color="tab:red", label='Blade')
                #     ax.plot(x_defl - yps - blade.nx[0] + lateral_clearance, z_defl, color="tab:red")
                #     ax.plot(blade.nx+ yss - blade.nx[0] + lateral_clearance, blade.nz, color="tab:blue", label='Undeflected blade')
                #     ax.plot(blade.nx- yps - blade.nx[0] + lateral_clearance, blade.nz, color="tab:blue")
                #     ax.set_xlim([0,50])
                #     ax.set_ylim([-0,110])
                #     ax.set_xlabel('m', weight='bold')
                #     ax.set_ylabel('m', weight='bold')
                #     plt.grid(color=[0.8,0.8,0.8], linestyle='--')
                #     plt.subplots_adjust(bottom = 0.15, left = 0.2)
                #     ax.legend()
                #     plt.savefig('/Users/pbortolo/work/9_Presentations/FY22/BAR/RailReview/13deg_85m.pdf')
                #     plt.show()
                #     exit()

                # Derailing reaction force on root node
                #  - Lateral force on wheels (multiply by 0.5 for 2 wheel sets)
                #  - Moment around axis perpendicular to ground
                RF_derailH[k] = (
                    0.5 * np.sqrt(forces_rxn.Fx**2 + forces_rxn.Fz**2) + np.abs(forces_rxn.Myy) / flatcar_tc_length
                )

                # Element shear and bending, one per element, which are already in principle directions in Hansen's notation
                # Zero-ing out axial stress as there shouldn't be any for pure beam bending
                Fz = np.r_[-forces.Nx[0, 0], forces.Nx[0, 1::2]]
                M1 = np.r_[-forces.Myy[0, 0], forces.Myy[0, 1::2]]
                M2 = np.r_[forces.Mzz[0, 0], -forces.Mzz[0, 1::2]]
                if PBEAM:
                    M1, M2 = rotate(M1, M2)

                # Compute strain at the two points: pressure/suction side extremes
                strainPS[:, k] = -(
                    M1 / EI11 * ps2 - M2 / EI22 * ps1 + Fz / EA
                )  # negative sign because Hansen c3 is opposite of Precomp z
                strainSS[:, k] = -(M1 / EI11 * ss2 - M2 / EI22 * ss1 + Fz / EA)

            if optFlag:
                # First constraint is compliance with outer boundary
                cboundary_outer = np.maximum(r_check_outer, 0)
                cboundary_inner = np.maximum(r_check_inner, 0)

                # Second constraint is reaction forces for derailment:
                crxn = np.maximum(RF_derailH - (0.5 * mass_car * gravity) / max_LV, 0.0)

                # Third constraint is keeping the strains reasonable
                cstrainPS = np.maximum(np.abs(strainPS) - max_strains, 0.0)
                cstrainSS = np.maximum(np.abs(strainSS) - max_strains, 0.0)

                # Accumulate constraints
                cons = np.array(
                    [
                        np.sum(cboundary_outer),
                        np.sum(cboundary_inner),
                        np.sum(crxn),
                        np.sum(cstrainPS),
                        np.sum(cstrainSS),
                    ]
                )
                # cons = np.array([np.sum(cboundary_outer), np.sum(crxn), np.sum(cstrainPS), np.sum(cstrainSS)])

                return -cons
            else:
                return RF_derailH, strainPS, strainSS

        # Initiliaze scipy minimization: Minimize force applied that keeps blade within all constraints
        const = {}
        const["type"] = "ineq"
        const["fun"] = run_hcurve

        npts = 2 * (self.n_span - 1)
        bounds = [(-max_rot, max_rot), (-max_rot, max_rot)]
        for i in range(npts):
            bounds.append((0, 1e2))
        # bounds = [(-1e2, 1e2)] * (npts + 2)
        x0 = np.r_[result.x, 0.0 * np.ones(npts)]
        result = minimize(
            lambda x: np.sum(np.abs(x[2:])),
            # get_max_force_h,
            x0,
            method="slsqp",
            bounds=bounds,
            tol=1e-3,
            constraints=const,
            options={"maxiter": 100},
        )

        print("Rail transport module convergence: ", result.success)
        # Evaluate optimized solution
        RF_derailH, strainPS, strainSS = run_hcurve(result.x, optFlag=False, final=True)

        # Express derailing force as a constraint
        if _8axle:
            outputs["constr_LV_8axle_horiz"] = RF_derailH / (0.5 * mass_car * gravity) / max_LV
        else:
            outputs["constr_LV_4axle_horiz"] = RF_derailH / (0.5 * mass_car * gravity) / max_LV

        # Strain constraint outputs
        outputs["constr_strainPS"] = np.max(np.abs(strainPS) / max_strains, axis=1)
        outputs["constr_strainSS"] = np.max(np.abs(strainSS) / max_strains, axis=1)
        # else:
        #     outputs['LV_constraint_4axle_horiz'] = 2.
        #     outputs['LV_constraint_8axle_horiz'] = 2.
        #     outputs['constr_strainPS']           = 2. * np.ones([npts,2])
        #     outputs['constr_strainSS']           = 2. * np.ones([npts,2])
        #     print('The optimization cannot satisfy the blade rail transport constraints.')

        """
        # ------- Vertical hills/sag using best attachment points
        # Set up Frame3DD blade for vertical analysis

        # Node location starting points
        nodes   = pyframe3dd.NodeData(inode, x_ref, r_curveV+y_ref, z_ref, rad)

        # Compute node radii starting points to determine if within clearance boundary
        r_blade = np.sqrt(nodes.y**2 + nodes.z**2)

        # Initialize frame3dd object
        blade = pyframe3dd.Frame(nodes, reactions, elements, options)

        # Hill
        blade_ymax        = y_ref + xte
        r_envelopeV       = r_curveV + vertical_clearance
        r_envelopeV_outer = r_envelopeV - blade_ymax
        node_dr           = np.minimum(r_envelopeV_outer - r_blade, 0)
        node_dy           = node_dr*np.cos(arcsV)
        node_dz           = node_dr*np.sin(arcsV)

        # Load case 1: gravity + hill
        dx = dM = np.zeros(ireact.size)
        load1 = pyframe3dd.StaticLoadCase(gx, gy, gz)
        load1.changePrescribedDisplacements(ireact+1, dx, node_dy[ireact], node_dz[ireact], dM, dM, dM)

        # Sag
        blade_ymin        = y_ref - xte
        r_envelopeV       = r_curveV - vertical_clearance
        r_envelopeV_inner = r_envelopeV.min() + blade_ymin
        node_dr           = np.maximum(r_envelopeV_inner - r_blade, 0)
        node_dy           = node_dr*np.cos(arcsV)
        node_dz           = node_dr*np.sin(arcsV)

        # Load case 2: gravity + sag
        load2 = pyframe3dd.StaticLoadCase(gx, gy, gz)
        load2.changePrescribedDisplacements(ireact+1, dx, node_dy[ireact], node_dz[ireact], dM, dM, dM)

        # Store these load cases and run
        blade.addLoadCase(load1)
        blade.addLoadCase(load2)
        #blade.write('blade.3dd')
        displacements, forces, forces_rxn, internalForces, mass, modal = blade.run()

        # Reaction forces for derailment:
        #  - Lateral force on wheels (multiply by 0.5 for 2 wheel sets)
        #  - Moment around axis perpendicular to ground
        # Should have 2 cases X 3 rxn nodes
        RF_derailV = -0.5*forces_rxn.Fy - forces_rxn.Mxx/flatcar_tc_length

        # Loop over hill & sag cases, then take worst strain case
        strainLE = np.zeros((self.n_span, 2))
        strainTE = np.zeros((self.n_span, 2))
        for k in range(2):
            # Element shear and bending, one per element, with conversion to profile c.s. using Hansen's notation
            # Zero-ing out axial stress as there shouldn't be any for pure beam bending
            Fz = 0.0 #np.r_[-forces.Nx[ 0,0],  forces.Nx[ 0, 1::2]]
            M1 = np.r_[-forces.Myy[k, 0],  forces.Myy[k, 1::2]]
            M2 = np.r_[ forces.Mzz[k, 0], -forces.Mzz[k, 1::2]]

            # compute strain at the two points
            strainLE[:,k] = -(M1/EI11*le2 - M2/EI22*le1 + Fz/EA)
            strainTE[:,k] = -(M1/EI11*te2 - M2/EI22*te1 + Fz/EA)

        # Find best points for middle reaction and formulate as constraints
        constr_derailV_8axle = (np.abs(RF_derailV.T) / (0.5 * mass_car_8axle * gravity)) / max_LV
        constr_derailV_4axle = (np.abs(RF_derailV.T) / (0.5 * mass_car_4axle * gravity)) / max_LV

        outputs['constr_LV_4axle_vert'] = constr_derailV_4axle[0,:]
        outputs['constr_LV_8axle_vert'] = constr_derailV_8axle[0,:]

        # Strain constraint outputs
        outputs['constr_strainLE'] = strainLE / max_strains
        outputs['constr_strainTE'] = strainTE / max_strains
        """



import os
import sys

import pytest

testpath = (
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    + os.path.sep
    + "wisdem"
    + os.path.sep
    + "test"
    + os.path.sep
)

if __name__ == "__main__":
    sys.exit(pytest.main([testpath]))



"""
This Jacket analysis and design module largely follows the work presented in:
"A systematic approach to offshore wind turbine jacket predesign and optimization:
geometry, cost, and surrogate structural code check models"
by Jan Häfele, Rick R. Damiani, Ryan N. King, Cristian G. Gebhardt, and Raimund Rolfes
Accessible via: https://wes.copernicus.org/articles/3/553/2018/
"""

import numpy as np
import openmdao.api as om

import wisdem.pyframe3dd.pyframe3dd as pyframe3dd
import wisdem.commonse.manufacturing as manu
import wisdem.commonse.cross_sections as cs
import wisdem.commonse.utilization_dnvgl as util_dnvgl
import wisdem.commonse.utilization_constraints as util_con
from wisdem.commonse import NFREQ, RIGID, gravity


class PreDiscretization(om.ExplicitComponent):
    """
    Process some of the tower YAML inputs.

    Parameters
    ----------
    tower_foundation_height : float, [m]
        Starting z-coordinate value of the tower with 0 at the water line
    transition_piece_height : float, [m]
        Point mass height of transition piece above water line
    transition_piece_mass : float, [kg]
        Point mass of transition piece
    transition_piece_cost : float, [USD]
        Cost of transition piece

    Returns
    -------
    transition_piece_height : float, [m]
        Point mass height of transition piece above water line
    joint1 : numpy array[3], [m]
        Global dimensional coordinates (x-y-z) for bottom node of member
    joint2 : numpy array[3], [m]
        Global dimensional coordinates (x-y-z) for top node of member
    suctionpile_depth : float, [m]
        Depth of monopile below sea floor
    bending_height : float, [m]
        Length of monopile above mudline subject to bending

    """

    def setup(self):
        self.add_input("tower_foundation_height", val=0.0, units="m")
        self.add_input("tower_base_diameter", val=0.0, units="m")
        self.add_input("transition_piece_mass", 0.0, units="kg")

        self.add_output("transition_piece_height", 0.0, units="m")
        self.add_output("transition_node", val=np.zeros(3), units="m")
        self.add_output("transition_piece_I", np.zeros(6), units="kg*m**2")

    def compute(self, inputs, outputs):
        # Unpack values
        fh_tow = inputs["tower_foundation_height"]
        m_trans = inputs["transition_piece_mass"]
        d_trans = inputs["tower_base_diameter"]

        outputs["transition_piece_height"] = fh_tow
        outputs["transition_node"] = np.r_[0.0, 0.0, fh_tow]

        # Mass properties for transition piece and gravity foundation
        r_trans = 0.5 * d_trans
        I_trans = m_trans * r_trans**2.0 * np.r_[0.5, 0.5, 1.0, np.zeros(3)]  # shell
        outputs["transition_piece_I"] = I_trans


class ComputeJacketNodes(om.ExplicitComponent):
    """"""

    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        mod_opt = self.options["modeling_options"]
        n_legs = mod_opt["WISDEM"]["FixedBottomSE"]["n_legs"]
        n_bays = mod_opt["WISDEM"]["FixedBottomSE"]["n_bays"]

        self.add_input("height", val=70.0, units="m")
        self.add_input("r_head", val=6.0, units="m")
        self.add_input("foot_head_ratio", val=1.5)
        self.add_input("bay_spacing", val=np.linspace(0.1, 0.9, n_bays + 1))
        self.add_input("tower_base_diameter", val=0.0, units="m")

        self.add_output("r_foot", val=10.0, units="m")
        self.add_output("leg_nodes", val=np.zeros((n_legs, n_bays + 2, 3)), units="m")
        self.add_output("bay_nodes", val=np.zeros((n_legs, n_bays + 1, 3)), units="m")
        self.add_output("constr_diam_consistency", val=0.0)

    def compute(self, inputs, outputs):
        mod_opt = self.options["modeling_options"]
        n_legs = mod_opt["WISDEM"]["FixedBottomSE"]["n_legs"]
        n_bays = mod_opt["WISDEM"]["FixedBottomSE"]["n_bays"]

        ratio = inputs["foot_head_ratio"]
        r_head = inputs["r_head"]
        outputs["r_foot"] = r_foot = r_head * ratio
        height = inputs["height"]

        leg_spacing = np.linspace(0, 1.0, n_bays + 2)
        z_leg = leg_spacing * height
        x_leg = leg_spacing * (r_head - r_foot) + r_foot
        for i in range(n_legs):
            outputs["leg_nodes"][i, :, 0] = x_leg * np.cos(i / n_legs * 2 * np.pi)
            outputs["leg_nodes"][i, :, 1] = x_leg * np.sin(i / n_legs * 2 * np.pi)
            outputs["leg_nodes"][i, :, 2] = z_leg

        z_bay = inputs["bay_spacing"] * height
        x_bay = inputs["bay_spacing"] * (r_head - r_foot) + r_foot
        for i in range(n_legs):
            outputs["bay_nodes"][i, :, 0] = x_bay * np.cos(i / n_legs * 2 * np.pi)
            outputs["bay_nodes"][i, :, 1] = x_bay * np.sin(i / n_legs * 2 * np.pi)
            outputs["bay_nodes"][i, :, 2] = z_bay

        outputs["constr_diam_consistency"] = inputs["tower_base_diameter"] / r_head


class ComputeFrame3DD(om.ExplicitComponent):
    """
    Now that we have modal information of the jacket structure, we construct the
    Frame3DD problem to solve for the structural properties of the jacket under
    the designated loading conditions.

    This is a lengthy process that requires creating a singular nodal array,
    member information for each brace and leg segment, and bringing in the
    loading from the tower.

    There are n_legs "ghost" members at the top of the jacket structure to connect
    to a ghost node that receives the turbine F and M values. These members are
    rigid and are needed to transmit the loads, but are not included in the mass,
    stress, and buckling calculations.
    """

    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        mod_opt = self.options["modeling_options"]
        n_dlc = mod_opt["WISDEM"]["n_dlc"]
        n_legs = mod_opt["WISDEM"]["FixedBottomSE"]["n_legs"]
        n_bays = mod_opt["WISDEM"]["FixedBottomSE"]["n_bays"]
        x_mb = mod_opt["WISDEM"]["FixedBottomSE"]["mud_brace"]
        n_mat = mod_opt["materials"]["n_mat"]

        self.add_input("leg_nodes", val=np.zeros((n_legs, n_bays + 2, 3)), units="m")
        self.add_input("bay_nodes", val=np.zeros((n_legs, n_bays + 1, 3)), units="m")
        self.add_input("leg_diameter", val=1.4, units="m")
        self.add_input("leg_thickness", val=0.1, units="m")
        self.add_input("brace_diameters", val=np.ones((n_bays)), units="m")
        self.add_input("brace_thicknesses", val=np.ones((n_bays)) * 0.1, units="m")

        self.add_discrete_input("material_names", val=n_mat * [""])
        self.add_input("E_mat", val=np.zeros([n_mat, 3]), units="Pa")
        self.add_input("G_mat", val=np.zeros([n_mat, 3]), units="Pa")
        self.add_input("rho_mat", val=np.zeros(n_mat), units="kg/m**3")
        self.add_input("sigma_y_mat", val=np.zeros(n_mat), units="Pa")

        self.add_input("turbine_F", np.zeros((3, n_dlc)), units="N")
        self.add_input("turbine_M", np.zeros((3, n_dlc)), units="N*m")
        self.add_input("transition_piece_mass", 0.0, units="kg")
        self.add_input("transition_piece_I", np.zeros(6), units="kg*m**2")
        self.add_input("gravity_foundation_mass", 0.0, units="kg")
        self.add_input("gravity_foundation_I", np.zeros(6), units="kg*m**2")
        self.add_input("tower_mass", val=0.0, units="kg")

        # For modal analysis only (loads captured in turbine_F & turbine_M)
        self.add_input("turbine_mass", val=0.0, units="kg")
        self.add_input("turbine_cg", val=np.zeros(3), units="m")
        self.add_input("turbine_I", np.zeros(6), units="kg*m**2")

        n_node = 3 * n_legs * (n_bays + 1) + 1
        self.add_output("jacket_nodes", np.zeros((n_node, 3)), units="m")
        self.add_output("leg_indices", np.zeros(n_legs))

        n_elem = 2 * (n_legs * (n_bays + 1)) + 4 * (n_legs * n_bays) + int(x_mb) * n_legs + n_legs

        self.add_output("jacket_elem_N", np.zeros((n_elem, 2)))
        self.add_output("jacket_elem_L", np.zeros(n_elem), units="m")
        self.add_output("jacket_elem_D", np.zeros(n_elem), units="m")
        self.add_output("jacket_elem_t", np.zeros(n_elem), units="m")
        self.add_output("jacket_elem_A", np.zeros(n_elem), units="m**2")
        self.add_output("jacket_elem_Asx", np.zeros(n_elem), units="m**2")
        self.add_output("jacket_elem_Asy", np.zeros(n_elem), units="m**2")
        self.add_output("jacket_elem_Ixx", np.zeros(n_elem), units="kg*m**2")
        self.add_output("jacket_elem_Iyy", np.zeros(n_elem), units="kg*m**2")
        self.add_output("jacket_elem_J0", np.zeros(n_elem), units="kg*m**2")
        self.add_output("jacket_elem_E", np.zeros(n_elem), units="Pa")
        self.add_output("jacket_elem_G", np.zeros(n_elem), units="Pa")
        self.add_output("jacket_elem_rho", np.zeros(n_elem), units="kg/m**3")
        self.add_output("jacket_elem_sigma_y", np.zeros(n_elem), units="Pa")
        self.add_output("jacket_elem_qdyn", np.zeros((n_elem, n_dlc)), units="Pa")
        self.add_output("jacket_mass", 0.0, units="kg")
        self.add_output("structural_mass", val=0.0, units="kg")

        self.add_output("jacket_base_F", np.zeros((3, n_dlc)), units="N")
        self.add_output("jacket_base_M", np.zeros((3, n_dlc)), units="N*m")
        self.add_output("jacket_Fz", np.zeros((n_elem, n_dlc)), units="N")
        self.add_output("jacket_Vx", np.zeros((n_elem, n_dlc)), units="N")
        self.add_output("jacket_Vy", np.zeros((n_elem, n_dlc)), units="N")
        self.add_output("jacket_Mxx", np.zeros((n_elem, n_dlc)), units="N*m")
        self.add_output("jacket_Myy", np.zeros((n_elem, n_dlc)), units="N*m")
        self.add_output("jacket_Mzz", np.zeros((n_elem, n_dlc)), units="N*m")

        # Frequencies
        self.add_output("f1", val=0.0, units="Hz")
        self.add_output("f2", val=0.0, units="Hz")
        self.add_output("structural_frequencies", np.zeros(NFREQ), units="Hz")

        self.add_output("jacket_deflection", np.zeros((n_node, n_dlc)), units="m")
        self.add_output("top_deflection", np.zeros(n_dlc), units="m")

        self.idx = 0

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        mod_opt = self.options["modeling_options"]
        n_dlc = mod_opt["WISDEM"]["n_dlc"]
        n_legs = mod_opt["WISDEM"]["FixedBottomSE"]["n_legs"]
        n_bays = mod_opt["WISDEM"]["FixedBottomSE"]["n_bays"]
        x_mb = mod_opt["WISDEM"]["FixedBottomSE"]["mud_brace"]
        material_name = mod_opt["WISDEM"]["FixedBottomSE"]["material"]

        leg_nodes = inputs["leg_nodes"]
        bay_nodes = inputs["bay_nodes"]

        # Add center of x joint nodes, in between bay members.
        x_nodes = np.zeros((n_legs, n_bays, 3))
        for jdx in range(n_bays):
            for idx in range(n_legs):
                n1 = bay_nodes[idx, jdx]
                n2 = bay_nodes[(idx + 1) % n_legs, (jdx + 1) % (n_bays + 1)]

                n3 = bay_nodes[(idx + 1) % n_legs, jdx]
                n4 = bay_nodes[idx, (jdx + 1) % (n_bays + 1)]

                # Filter out division by 0 warnings
                with np.errstate(divide="ignore", invalid="ignore"):
                    alpha = (n4 - n2) / ((n1 - n2) - (n3 - n4))
                alpha = alpha[0]
                new_node = n4 + alpha * (n3 - n4)
                x_nodes[idx, jdx, :] = new_node

        # Add ghost node for transition piece
        ghost_nodes = np.mean(leg_nodes[:, -1, :], axis=0)
        ghost_nodes[2] += 2.0  # add two meters in the z-direction
        ghost_nodes = ghost_nodes.reshape(1, 1, 3)

        # Concatenate all nodal information together
        xyz = np.vstack(
            (leg_nodes.reshape(-1, 3), bay_nodes.reshape(-1, 3), x_nodes.reshape(-1, 3), ghost_nodes.reshape(-1, 3))
        )
        n = xyz.shape[0]
        node_indices = np.arange(1, n + 1)
        r = np.zeros(n)
        nodes = pyframe3dd.NodeData(node_indices, xyz[:, 0], xyz[:, 1], xyz[:, 2], r)
        outputs["jacket_nodes"] = xyz

        # Create arrays to later reference the indices for each node. Needed because
        # Frame3DD expects a singular nodal array but we want to keep information
        # about which nodes belong to the legs, bays, or ghost nodes.
        leg_indices = node_indices[: leg_nodes.size // 3].reshape((n_legs, n_bays + 2))
        bay_indices = node_indices[leg_nodes.size // 3 : leg_nodes.size // 3 + bay_nodes.size // 3].reshape(
            (n_legs, n_bays + 1)
        )
        x_indices = node_indices[leg_nodes.size // 3 + bay_nodes.size // 3 : -1].reshape((n_legs, n_bays))
        ghost_indices = np.atleast_2d(node_indices[-1])

        rnode = np.array(leg_indices[:, 0], dtype=np.int_)
        kx = ky = kz = ktx = kty = ktz = np.array([RIGID] * len(rnode))
        reactions = pyframe3dd.ReactionData(rnode, kx, ky, kz, ktx, kty, ktz, rigid=RIGID)
        outputs["leg_indices"] = rnode
        # ------ frame element data ------------

        self.num_elements = 0
        self.N1 = []
        self.N2 = []
        self.L = []
        self.D = []
        self.t = []
        self.Area = []
        self.Asx = []
        self.Asy = []
        self.J0 = []
        self.Ixx = []
        self.Iyy = []
        self.vol = []

        # Helper function to add an element to all relevant lists.
        def add_element(n1_nodes, n1_indices, n2_nodes, n2_indices, itube, idx1, idx2, idx3, idx4):
            n1 = n1_nodes[idx1, idx2]
            n2 = n2_nodes[idx3, idx4]
            self.N1.append(n1_indices[idx1, idx2])
            self.N2.append(n2_indices[idx3, idx4])
            length = np.linalg.norm(n2 - n1)
            self.L.append(length)
            self.D.append(itube.D)
            self.t.append(itube.t)
            self.Area.append(itube.Area)
            self.Asx.append(itube.Asx)
            self.Asy.append(itube.Asy)
            self.J0.append(itube.J0)
            self.Ixx.append(itube.Ixx)
            self.Iyy.append(itube.Iyy)
            self.vol.append(itube.Area * length)
            self.num_elements += 1

        # Naive for loops to make sure we get indexing right.
        # Could vectorize later as needed.

        # Add leg members.
        for jdx in range(n_bays + 1):
            itube = cs.Tube(float(inputs["leg_diameter"][0]), float(inputs["leg_thickness"][0]))
            for idx in range(n_legs):
                add_element(leg_nodes, leg_indices, bay_nodes, bay_indices, itube, idx, jdx, idx, jdx)
                add_element(bay_nodes, bay_indices, leg_nodes, leg_indices, itube, idx, jdx, idx, jdx + 1)

        # Add brace (x-joint) members
        for jdx in range(n_bays):
            itube = cs.Tube(inputs["brace_diameters"][jdx], inputs["brace_thicknesses"][jdx])
            for idx in range(n_legs):
                add_element(bay_nodes, bay_indices, x_nodes, x_indices, itube, idx, jdx, idx, jdx)
                add_element(
                    x_nodes,
                    x_indices,
                    bay_nodes,
                    bay_indices,
                    itube,
                    idx,
                    jdx,
                    (idx + 1) % n_legs,
                    (jdx + 1) % (n_bays + 1),
                )

                add_element(bay_nodes, bay_indices, x_nodes, x_indices, itube, (idx + 1) % n_legs, jdx, idx, jdx)
                add_element(
                    x_nodes,
                    x_indices,
                    bay_nodes,
                    bay_indices,
                    itube,
                    idx,
                    jdx,
                    idx,
                    (jdx + 1) % (n_bays + 1),
                )

        # Add mud brace if boolean True
        if x_mb:
            itube = cs.Tube(inputs["brace_diameters"][0], inputs["brace_thicknesses"][0])
            for idx in range(n_legs):
                add_element(bay_nodes, bay_indices, bay_nodes, bay_indices, itube, idx, 0, (idx + 1) % n_legs, 0)

        # Add ghost point where we add the turbine_F and M as well as transition mass
        itube = cs.Tube(1.0e-2, 1.0e-3)
        for idx in range(n_legs):
            add_element(leg_nodes, leg_indices, ghost_nodes, ghost_indices, itube, idx, -1, 0, 0)

        # Grab material information; generally steel will be used
        imat = discrete_inputs["material_names"].index(material_name)
        E = [np.mean(inputs["E_mat"][imat])] * self.num_elements
        G = [np.mean(inputs["G_mat"][imat])] * self.num_elements
        rho = [inputs["rho_mat"][imat]] * self.num_elements

        # Convert all lists to arrays
        Area = np.squeeze(np.array(self.Area, dtype=np.float_))
        Asx = np.squeeze(np.array(self.Asx, dtype=np.float_))
        Asy = np.squeeze(np.array(self.Asy, dtype=np.float_))
        J0 = np.squeeze(np.array(self.J0, dtype=np.float_))
        Ixx = np.squeeze(np.array(self.Ixx, dtype=np.float_))
        Iyy = np.squeeze(np.array(self.Iyy, dtype=np.float_))
        L = np.squeeze(np.array(self.L, dtype=np.float_))
        D = np.squeeze(np.array(self.D, dtype=np.float_))
        t = np.squeeze(np.array(self.t, dtype=np.float_))
        E = np.squeeze(np.array(E, dtype=np.float_))
        G = np.squeeze(np.array(G, dtype=np.float_))
        rho = np.squeeze(np.array(rho, dtype=np.float_))
        N1 = self.N1
        N2 = self.N2

        # Modify last n_legs elements to make them rigid due to the ghost node
        E[-n_legs:] *= 1.0e8
        G[-n_legs:] *= 1.0e8
        rho[-n_legs:] = 1.0e-2

        outputs["jacket_elem_N"] = connect_mat = np.c_[N1, N2] - 1  # Storing as 0-indexed
        outputs["jacket_elem_L"] = L
        outputs["jacket_elem_D"] = D
        outputs["jacket_elem_t"] = t
        outputs["jacket_elem_A"] = Area
        outputs["jacket_elem_Asx"] = Asx
        outputs["jacket_elem_Asy"] = Asy
        outputs["jacket_elem_Ixx"] = Ixx
        outputs["jacket_elem_Iyy"] = Iyy
        outputs["jacket_elem_J0"] = J0
        outputs["jacket_elem_E"] = E
        outputs["jacket_elem_G"] = G
        outputs["jacket_elem_rho"] = rho
        outputs["jacket_elem_sigma_y"] = inputs["sigma_y_mat"][imat]
        outputs["jacket_elem_qdyn"] = 1.0e2  # hardcoded value for now
        outputs["jacket_elem_sigma_y"][-n_legs:] *= 1e6
        outputs["jacket_elem_qdyn"][-n_legs:] *= 1e4

        element = np.arange(1, self.num_elements + 1)
        roll = np.zeros(self.num_elements)

        elements = pyframe3dd.ElementData(element, N1, N2, Area, Asx, Asy, J0, Ixx, Iyy, E, G, roll, rho)

        # Populate mass and cost outputs
        # TODO: Is there an "outfitting" factor for jackets.  Seems like it would be much smaller than monopiles
        outputs["jacket_mass"] = np.sum(Area[:-n_legs] * rho[:-n_legs] * L[:-n_legs])
        outputs["structural_mass"] = outputs["jacket_mass"] + inputs["tower_mass"]

        # ------ options ------------
        dx = -1.0
        options = pyframe3dd.Options(
            mod_opt["WISDEM"]["FixedBottomSE"]["frame3dd"]["shear"],
            mod_opt["WISDEM"]["FixedBottomSE"]["frame3dd"]["geom"],
            dx,
        )
        # -----------------------------------

        # initialize frame3dd object
        self.frame = pyframe3dd.Frame(nodes, reactions, elements, options)

        if mod_opt["WISDEM"]["FixedBottomSE"]["save_truss_figures"]:
            if not self.under_approx:
                self.frame.draw(savefig=True, fig_idx=self.idx)
                self.idx += 1

        # ------- enable dynamic analysis ----------
        Mmethod = 1
        lump = 0
        shift = 0.0
        # Run twice the number of modes to ensure that we can ignore the torsional modes and still get the desired number of fore-aft, side-side modes
        self.frame.enableDynamics(2 * NFREQ, Mmethod, lump, 1e-4, shift)
        # ----------------------------

        # ------ static load case 1 ------------
        # gravity in the X, Y, Z, directions (global)
        gx = 0.0
        gy = 0.0
        gz = -gravity

        for k in range(n_dlc):
            load_obj = pyframe3dd.StaticLoadCase(gx, gy, gz)

            # Prepare point forces at transition node
            turb_F = inputs["turbine_F"][:, k]
            turb_M = inputs["turbine_M"][:, k]
            load_obj.changePointLoads(
                np.array([n], dtype=np.int_),  # -1 b/c same reason as above
                np.array([turb_F[0]]).flatten(),
                np.array([turb_F[1]]).flatten(),
                np.array([turb_F[2]]).flatten(),
                np.array([turb_M[0]]).flatten(),
                np.array([turb_M[1]]).flatten(),
                np.array([turb_M[2]]).flatten(),
            )

            # # trapezoidally distributed loads
            # xx1 = xy1 = xz1 = np.zeros(ielem.size)
            # xx2 = xy2 = xz2 = 0.99 * L  # multiply slightly less than unity b.c. of precision
            # wx1 = inputs["jacket_elem_Px1"][:nelem, k]
            # wx2 = inputs["jacket_elem_Px2"][:nelem, k]
            # wy1 = inputs["jacket_elem_Py1"][:nelem, k]
            # wy2 = inputs["jacket_elem_Py2"][:nelem, k]
            # wz1 = inputs["jacket_elem_Pz1"][:nelem, k]
            # wz2 = inputs["jacket_elem_Pz2"][:nelem, k]
            # load_obj.changeTrapezoidalLoads(ielem, xx1, xx2, wx1, wx2, xy1, xy2, wy1, wy2, xz1, xz2, wz1, wz2)

            # Add the load case and run
            self.frame.addLoadCase(load_obj)

        # ------ add extra mass ------------
        # Prepare transition piece, and gravity foundation (if any applicable) for "extra node mass"
        # Turbine mass added for modal analysis only- gravity loads accounted for in point force
        m_trans = float(inputs["transition_piece_mass"][0])
        I_trans = inputs["transition_piece_I"].flatten()
        m_grav = float(inputs["gravity_foundation_mass"][0])
        I_grav = inputs["gravity_foundation_I"].flatten()
        m_turb = float(inputs["turbine_mass"][0])
        cg_turb = inputs["turbine_cg"].flatten()
        I_turb = inputs["turbine_I"].flatten()
        # Note, need len()-1 because Frame3DD crashes if mass add at end
        midx = np.array([n, n - 1, 1], dtype=np.int_)
        m_add = np.array([m_turb, m_trans, m_grav])
        mI = np.c_[I_turb, I_trans, I_grav]
        mrho = np.c_[cg_turb, np.zeros(3), np.zeros(3)]
        add_gravity = [False, True, True]
        self.frame.changeExtraNodeMass(
            midx,
            m_add,
            mI[0, :],
            mI[1, :],
            mI[2, :],
            mI[3, :],
            mI[4, :],
            mI[5, :],
            mrho[0, :],
            mrho[1, :],
            mrho[2, :],
            add_gravity,
        )

        # self.frame.write("jacket.3dd")
        # self.frame.draw()
        displacements, forces, reactions, internalForces, mass, modal = self.frame.run()

        # natural frequncies
        outputs["f1"] = modal.freq[0]
        outputs["f2"] = modal.freq[1]
        outputs["structural_frequencies"] = modal.freq[:NFREQ]

        # deflections due to loading (from cylinder top and wind/wave loads)
        outputs["jacket_deflection"] = np.sqrt(displacements.dx**2 + displacements.dy**2).T
        outputs["top_deflection"] = outputs["jacket_deflection"][-1, :]

        # Determine reaction forces
        outputs["jacket_base_F"] = -np.c_[
            reactions.Fx.sum(axis=1), reactions.Fy.sum(axis=1), reactions.Fz.sum(axis=1)
        ].T
        outputs["jacket_base_M"] = -np.c_[
            reactions.Mxx.sum(axis=1), reactions.Myy.sum(axis=1), reactions.Mzz.sum(axis=1)
        ].T

        for ic in range(n_dlc):
            # Forces and moments along the structure
            outputs["jacket_Fz"][:, ic] = forces.Nx[ic, 1::2]
            outputs["jacket_Vx"][:, ic] = -forces.Vz[ic, 1::2]
            outputs["jacket_Vy"][:, ic] = forces.Vy[ic, 1::2]
            outputs["jacket_Mxx"][:, ic] = -forces.Mzz[ic, 1::2]
            outputs["jacket_Myy"][:, ic] = forces.Myy[ic, 1::2]
            outputs["jacket_Mzz"][:, ic] = forces.Txx[ic, 1::2]


class JacketPost(om.ExplicitComponent):
    """
    This component computes the stress and buckling utilization values
    for the jacket structure based on the loading computed by Frame3DD.
    """

    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        mod_opt = self.options["modeling_options"]
        n_dlc = mod_opt["WISDEM"]["n_dlc"]
        n_legs = mod_opt["WISDEM"]["FixedBottomSE"]["n_legs"]
        n_bays = mod_opt["WISDEM"]["FixedBottomSE"]["n_bays"]
        x_mb = mod_opt["WISDEM"]["FixedBottomSE"]["mud_brace"]

        n_elem = 2 * (n_legs * (n_bays + 1)) + 4 * (n_legs * n_bays) + int(x_mb) * n_legs + n_legs

        self.add_input("jacket_elem_L", np.zeros(n_elem), units="m")
        self.add_input("jacket_elem_D", np.zeros(n_elem), units="m")
        self.add_input("jacket_elem_t", np.zeros(n_elem), units="m")
        self.add_input("jacket_elem_A", np.zeros(n_elem), units="m**2")
        self.add_input("jacket_elem_Asx", np.zeros(n_elem), units="m**2")
        self.add_input("jacket_elem_Asy", np.zeros(n_elem), units="m**2")
        self.add_input("jacket_elem_Ixx", np.zeros(n_elem), units="kg*m**2")
        self.add_input("jacket_elem_Iyy", np.zeros(n_elem), units="kg*m**2")
        self.add_input("jacket_elem_J0", np.zeros(n_elem), units="kg*m**2")
        self.add_input("jacket_elem_E", np.zeros(n_elem), units="Pa")
        self.add_input("jacket_elem_G", np.zeros(n_elem), units="Pa")
        self.add_input("jacket_elem_sigma_y", np.zeros(n_elem), units="Pa")
        self.add_input("jacket_elem_qdyn", np.zeros((n_elem, n_dlc)), units="Pa")

        # Processed Frame3DD/OpenFAST outputs
        self.add_input("jacket_Fz", np.ones((n_elem, n_dlc)), units="N")
        self.add_input("jacket_Vx", np.ones((n_elem, n_dlc)), units="N")
        self.add_input("jacket_Vy", np.ones((n_elem, n_dlc)), units="N")
        self.add_input("jacket_Mxx", np.ones((n_elem, n_dlc)), units="N*m")
        self.add_input("jacket_Myy", np.ones((n_elem, n_dlc)), units="N*m")
        self.add_input("jacket_Mzz", np.ones((n_elem, n_dlc)), units="N*m")

        # We don't care about the last n_leg members because they're connected
        # to the ghost node to transmit loads and masses.
        # Thus, the stress and buckling outputs are smaller dimensions
        # than the full n_elem.
        self.add_output("constr_stress", np.ones((n_elem - n_legs, n_dlc)))
        self.add_output("constr_shell_buckling", np.ones((n_elem - n_legs, n_dlc)))
        self.add_output("constr_global_buckling", np.ones((n_elem - n_legs, n_dlc)))

    def compute(self, inputs, outputs):
        # Unpack some variables
        mod_opt = self.options["modeling_options"]
        n_dlc = mod_opt["WISDEM"]["n_dlc"]
        n_legs = mod_opt["WISDEM"]["FixedBottomSE"]["n_legs"]
        gamma_f = mod_opt["WISDEM"]["FixedBottomSE"]["gamma_f"]
        gamma_m = mod_opt["WISDEM"]["FixedBottomSE"]["gamma_m"]
        gamma_n = mod_opt["WISDEM"]["FixedBottomSE"]["gamma_n"]
        gamma_b = mod_opt["WISDEM"]["FixedBottomSE"]["gamma_b"]

        d = inputs["jacket_elem_D"]
        t = inputs["jacket_elem_t"]
        h = inputs["jacket_elem_L"]
        Az = inputs["jacket_elem_A"]
        Asx = inputs["jacket_elem_Asx"]
        Jz = inputs["jacket_elem_J0"]
        Iyy = inputs["jacket_elem_Iyy"]
        sigy = inputs["jacket_elem_sigma_y"]
        E = inputs["jacket_elem_E"]
        G = inputs["jacket_elem_G"]
        qdyn = inputs["jacket_elem_qdyn"]
        r = 0.5 * d

        # Get loads from Framee3dd/OpenFAST
        Fz = inputs["jacket_Fz"]
        Vx = inputs["jacket_Vx"]
        Vy = inputs["jacket_Vy"]
        Mxx = inputs["jacket_Mxx"]
        Myy = inputs["jacket_Myy"]
        Mzz = inputs["jacket_Mzz"]

        M = np.sqrt(Mxx**2 + Myy**2)
        V = np.sqrt(Vx**2 + Vy**2)

        # See http://svn.code.sourceforge.net/p/frame3dd/code/trunk/doc/Frame3DD-manual.html#structuralmodeling
        # print(Fz.shape, Az.shape, M.shape, r.shape, Iyy.shape)
        axial_stress = Fz / Az[:, np.newaxis] + M * (r / Iyy)[:, np.newaxis]
        shear_stress = np.abs(Mzz) / (Jz * r)[:, np.newaxis] + V / Asx[:, np.newaxis]
        hoop_stress = -qdyn * ((r - 0.5 * t) / t)[:, np.newaxis]  # util_con.hoopStress(d, t, qdyn)
        outputs["constr_stress"] = util_con.TubevonMisesStressUtilization(
            axial_stress, hoop_stress, shear_stress, gamma_f * gamma_m * gamma_n, sigy.reshape((-1, 1))
        )[:-n_legs]

        # Use DNV-GL CP202 Method
        check = util_dnvgl.CylinderBuckling(h, d, t, E=E, G=G, sigma_y=sigy, gamma=gamma_f * gamma_b)
        for k in range(n_dlc):
            results = check.run_buckling_checks(
                Fz[:, k], M[:, k], axial_stress[:, k], hoop_stress[:, k], shear_stress[:, k]
            )

            outputs["constr_shell_buckling"][:, k] = results["Shell"][:-n_legs]
            outputs["constr_global_buckling"][:, k] = results["Global"][:-n_legs]


class JacketCost(om.ExplicitComponent):
    """
    Compute the jacket costs using textbook relations, much like the monopile calcs.

    """

    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        mod_opt = self.options["modeling_options"]
        n_legs = mod_opt["WISDEM"]["FixedBottomSE"]["n_legs"]
        n_bays = mod_opt["WISDEM"]["FixedBottomSE"]["n_bays"]
        x_mb = mod_opt["WISDEM"]["FixedBottomSE"]["mud_brace"]
        n_mat = mod_opt["materials"]["n_mat"]
        n_node = 3 * n_legs * (n_bays + 1) + 1
        n_elem = 2 * (n_legs * (n_bays + 1)) + 4 * (n_legs * n_bays) + int(x_mb) * n_legs + n_legs

        self.add_input("leg_nodes", val=np.zeros((n_legs, n_bays + 2, 3)), units="m")

        self.add_discrete_input("material_names", val=n_mat * [""])
        self.add_input("unit_cost_mat", val=np.zeros(n_mat), units="USD/kg")
        self.add_input("labor_cost_rate", 0.0, units="USD/min")
        self.add_input("painting_cost_rate", 0.0, units="USD/m/m")

        self.add_input("jacket_nodes", np.zeros((n_node, 3)), units="m")
        self.add_input("jacket_elem_N", np.zeros((n_elem, 2)))
        self.add_input("jacket_elem_L", np.zeros(n_elem), units="m")
        self.add_input("jacket_elem_D", np.zeros(n_elem), units="m")
        self.add_input("jacket_elem_t", np.zeros(n_elem), units="m")
        self.add_input("jacket_mass", 0.0, units="kg")

        self.add_input("tower_cost", val=0.0, units="USD")
        self.add_input("transition_piece_cost", 0.0, units="USD")

        self.add_output("labor_hours", 0.0, units="h")
        self.add_output("jacket_cost", 0.0, units="USD")
        self.add_output("structural_cost", val=0.0, units="USD")

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        mod_opt = self.options["modeling_options"]
        n_legs = mod_opt["WISDEM"]["FixedBottomSE"]["n_legs"]
        material_name = mod_opt["WISDEM"]["FixedBottomSE"]["material"]
        eps = 1e-8

        # Unpack inputs and ignore ghost nodes
        c_trans = inputs["transition_piece_cost"]
        leg_nodes = inputs["leg_nodes"]
        connect_mat = np.int_(inputs["jacket_elem_N"][:-n_legs, :])
        L = inputs["jacket_elem_L"][:-n_legs]
        D = inputs["jacket_elem_D"][:-n_legs].copy()
        t = inputs["jacket_elem_t"][:-n_legs]
        xyz = inputs["jacket_nodes"]
        m_total = inputs["jacket_mass"]
        n_edges = L.size
        N0 = connect_mat[:, 0]
        N1 = connect_mat[:, 1]

        # Compute costs based on "Optimum Design of Steel Structures" by Farkas and Jarmai
        imat = discrete_inputs["material_names"].index(material_name)
        k_m = inputs["unit_cost_mat"][imat]  # 1.1 # USD / kg carbon steel plate
        k_f = inputs["labor_cost_rate"]  # 1.0 # USD / min labor
        k_p = inputs["painting_cost_rate"]  # USD / m^2 painting
        k_e = 0.064  # Industrial electricity rate $/kWh https://www.eia.gov/electricity/monthly/epm_table_grapher.php?t=epmt_5_6_a
        e_f = 15.9  # Electricity usage kWh/kg for steel
        # e_fo = 26.9  # Electricity usage kWh/kg for stainless steel
        theta = 3  # Manufacturing difficulty factor

        # Prep vectors for elements and legs for some vector math
        edge_vec = xyz[N0, :] - xyz[N1, :]
        leg_vec = np.squeeze(leg_nodes[:, -1, :] - leg_nodes[:, 0, :])
        leg_L = np.linalg.norm(leg_vec, axis=1)

        # Get the angle of intersection between all edges (as vectors) and all legs (as vectors)
        vec_vals = np.dot(edge_vec, leg_vec.T) / np.outer(L, leg_L)
        leg_alpha = np.arccos(np.minimum(np.maximum(vec_vals, -1.0), 1.0))
        # If angle of intersection is close to 0 or 180, the edge is part of a leg
        tol = np.deg2rad(5)
        idx_leg = np.any((np.abs(leg_alpha) < tol) | (np.abs(leg_alpha - np.pi) < tol), axis=1)
        D[idx_leg] = 0.0  # No double-counting time for leg elements since single piece

        # Now determine which angle to use based on which leg a given edge node is on
        edge_alpha = 0.5 * np.pi * np.ones(n_edges)
        for k in range(n_legs):
            tol = 1e-2 * leg_L[k]
            sec1 = np.linalg.norm(np.squeeze(leg_nodes[k, -1, :])[np.newaxis, :] - xyz[N0, :], axis=1)
            sec2 = np.linalg.norm(xyz[N0, :] - np.squeeze(leg_nodes[k, 0, :])[np.newaxis, :], axis=1)
            on_leg_k = np.abs(leg_L[k] - sec1 - sec2) < tol
            edge_alpha[on_leg_k] = leg_alpha[on_leg_k, k]
        edge_alpha = np.minimum(edge_alpha, np.pi - edge_alpha) + eps

        # Run manufacturing time estimate functions
        weld_L = 2 * np.pi * D / np.sin(edge_alpha)  # Multiply by 2 for both ends
        n_pieces = n_edges - np.count_nonzero(D)
        t_cut = 2 * manu.steel_tube_cutgrind_time(theta, 0.5 * D, t, edge_alpha)  # Multiply by 2 for both ends
        t_weld = manu.steel_tube_welding_time(theta, n_pieces, m_total, weld_L, t)
        t_manufacture = t_cut + t_weld
        K_f = k_f * t_manufacture
        outputs["labor_hours"] = t_manufacture / 60.0
        
        # Cost step 5) Painting by surface area
        theta_p = 2
        K_p = k_p * theta_p * np.pi * np.sum(D[:-n_legs] * L[:-n_legs])

        # Material cost with waste fraction, but without outfitting,
        K_m = 1.21 * np.sum(k_m * m_total)

        # Electricity usage
        K_e = k_e * (e_f * m_total)  # + e_fo * (coeff - 1.0) * m_total

        # Assemble all costs for now
        tempSum = K_m + K_e + K_p + K_f

        # Capital cost share from BLS MFP by NAICS
        K_c = 0.118 * tempSum / (1.0 - 0.118)

        outputs["jacket_cost"] = K_c + tempSum + c_trans
        outputs["structural_cost"] = outputs["jacket_cost"] + inputs["tower_cost"]


# Assemble the system together in an OpenMDAO Group
class JacketSEProp(om.Group):
    """
    Group to contain all subsystems needed for jacket analysis and design.
    Can be used as a standalone or within the larger WISDEM stack.
    """

    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        modeling_options = self.options["modeling_options"]

        self.add_subsystem("pre", PreDiscretization(), promotes=["*"])
        self.add_subsystem("nodes", ComputeJacketNodes(modeling_options=modeling_options), promotes=["*"])
        self.add_subsystem("cost", JacketCost(modeling_options=modeling_options), promotes=["*"])


class JacketSEPerf(om.Group):
    """
    Group to contain all subsystems needed for jacket analysis and design.
    Can be used as a standalone or within the larger WISDEM stack.
    """

    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        modeling_options = self.options["modeling_options"]

        self.add_subsystem("frame3dd", ComputeFrame3DD(modeling_options=modeling_options), promotes=["*"])
        self.add_subsystem("post", JacketPost(modeling_options=modeling_options), promotes=["*"])


class JacketSE(om.Group):
    """
    Group to contain all subsystems needed for jacket analysis and design.
    Can be used as a standalone or within the larger WISDEM stack.
    """

    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        modeling_options = self.options["modeling_options"]

        self.add_subsystem("prop", JacketSEProp(modeling_options=modeling_options), promotes=["*"])
        self.add_subsystem("perf", JacketSEPerf(modeling_options=modeling_options), promotes=["*"])




import numpy as np

class Profile:
    """Defines the shape of an airfoil"""

    def __init__(self, xu, yu, xl, yl):
        """Constructor

        Parameters
        ----------
        xu : ndarray
            x coordinates for upper surface of airfoil
        yu : ndarray
            y coordinates for upper surface of airfoil
        xl : ndarray
            x coordinates for lower surface of airfoil
        yl : ndarray
            y coordinates for lower surface of airfoil

        Notes
        -----
        uses :ref:`airfoil-aligned coordinate system <blade_airfoil_coord>`.
        Nodes should be ordered from the leading edge toward the trailing edge.
        The leading edge can be located at any position and the
        chord may be any size, however the airfoil should be untwisted.
        Normalization to unit chord will happen internally.

        """

        # parse airfoil data
        xu = np.array(xu)
        yu = np.array(yu)
        xl = np.array(xl)
        yl = np.array(yl)

        # ensure leading edge at zero
        xu -= xu[0]
        xl -= xl[0]
        yu -= yu[0]
        yl -= yl[0]

        # ensure unit chord
        c = xu[-1] - xu[0]
        xu /= c
        xl /= c
        yu /= c
        yl /= c

        # interpolate onto common grid
        arc = np.linspace(0, np.pi, 100)
        self.x = 0.5 * (1 - np.cos(arc))  # cosine spacing
        self.yu = np.interp(self.x, xu, yu)
        self.yl = np.interp(self.x, xl, yl)

    @classmethod
    def initWithTEtoTEdata(cls, x, y):
        """Factory constructor for data points ordered from trailing edge to trailing edge.

        Parameters
        ----------
        x, y : ndarray, ndarray
            airfoil coordinates starting at trailing edge and
            ending at trailing edge, traversing airfoil in either direction.

        Returns
        -------
        profile : Profile
            initialized Profile object

        Notes
        -----
        It is not necessary to start and end at the same point
        for an airfoil with trailing edge thickness.
        Although, one point should be right at the nose.
        see also notes for :meth:`__init__`

        """

        # parse airfoil data
        x = np.array(x)
        y = np.array(y)

        # separate into 2 halves
        i = np.argmin(x)

        xu = x[i::-1]
        yu = y[i::-1]
        xl = x[i:]
        yl = y[i:]

        # check if coordinates were input in other direction
        if np.mean(y[0:i]) < np.mean(y[i:]):
            temp = yu
            yu = yl
            yl = temp

            temp = xu
            xu = xl
            xl = temp

        return cls(xu, yu, xl, yl)

    @classmethod
    def initWithLEtoLEdata(cls, x, y):
        """Factory constructor for data points ordered from leading edge to leading edge.

        Parameters
        ----------
        x, y : ndarray, ndarray
            airfoil coordinates starting at leading edge and
            ending at leading edge, traversing airfoil in either direction.

        Returns
        -------
        profile : Profile
            initialized Profile object

        Notes
        -----
        x,y data must start and end at the same point.
        see also notes for :meth:`__init__`

        """

        # parse airfoil data
        x = np.array(x)
        y = np.array(y)

        # separate into 2 halves
        for i in range(len(x)):
            if x[i + 1] <= x[i]:
                iuLast = i
                ilLast = i
                if x[i + 1] == x[i]:  # blunt t.e.
                    ilLast = i + 1  # stop at i+1
                break

        xu = x[: iuLast + 1]
        yu = y[: iuLast + 1]
        xl = x[-1 : ilLast - 1 : -1]
        yl = y[-1 : ilLast - 1 : -1]

        # check if coordinates were input in other direction
        if y[1] < y[0]:
            temp = yu
            yu = yl
            yl = temp

            temp = xu
            xu = xl
            xl = temp

        return cls(xu, yu, xl, yl)

    @staticmethod
    def initFromPreCompFile(precompProfileFile):
        """Construct profile from PreComp formatted file

        Parameters
        ----------
        precompProfileFile : str
            path/name of file

        Returns
        -------
        profile : Profile
            initialized Profile object

        """

        return Profile.initFromFile(precompProfileFile, 4, True)

    @staticmethod
    def initFromFile(filename, numHeaderlines, LEtoLE):
        """Construct profile from a generic form text file (see Notes)

        Parameters
        ----------
        filename : str
            name/path of input file
        numHeaderlines : int
            number of header rows in input file
        LEtoLE : boolean
            True if data is ordered from leading-edge to leading-edge
            False if from trailing-edge to trailing-edge

        Returns
        -------
        profile : Profile
            initialized Profile object

        Notes
        -----
        file should be of the form:

        header row
        header row
        x1 y1
        x2 y2
        x3 y3
        .  .
        .  .
        .  .

        where any number of header rows can be used.

        """

        # open file
        f = open(filename, "r")

        # skip through header
        for i in range(numHeaderlines):
            f.readline()

        # loop through
        x = []
        y = []

        for line in f:
            if not line.strip():
                break  # break if empty line
            data = line.split()
            x.append(float(data[0]))
            y.append(float(data[1]))

        f.close()

        # close nose if LE to LE
        if LEtoLE:
            x.append(x[0])
            y.append(y[0])
            return Profile.initWithLEtoLEdata(x, y)

        else:
            return Profile.initWithTEtoTEdata(x, y)

    def _preCompFormat(self):
        """
        docstring
        """

        # check if they share a common trailing edge point
        te_same = self.yu[-1] == self.yl[-1]

        # count number of points
        nu = len(self.x)
        if te_same:
            nu -= 1
        nl = len(self.x) - 1  # they do share common leading-edge
        n = nu + nl

        # initialize
        x = np.zeros(n)
        y = np.zeros(n)

        # leading edge round to leading edge
        x[0:nu] = self.x[0:nu]
        y[0:nu] = self.yu[0:nu]
        x[nu:] = self.x[:0:-1]
        y[nu:] = self.yl[:0:-1]

        return x, y

    def locationOfMaxThickness(self):
        """Find location of max airfoil thickness

        Returns
        -------
        x : float
            x location of maximum thickness
        yu : float
            upper surface y location of maximum thickness
        yl : float
            lower surface y location of maximum thickness

        Notes
        -----
        uses :ref:`airfoil-aligned coordinate system <blade_airfoil_coord>`

        """

        idx = np.argmax(self.yu - self.yl)
        return (self.x[idx], self.yu[idx], self.yl[idx])

    def blend(self, other, weight):
        """Blend this profile with another one with the specified weighting.

        Parameters
        ----------
        other : Profile
            another Profile to blend with
        weight : float
            blending parameter between 0 and 1.  0 returns self, whereas 1 returns other.

        Returns
        -------
        profile : Profile
            a blended profile

        """

        # blend coordinates
        yu = self.yu + weight * (other.yu - self.yu)
        yl = self.yl + weight * (other.yl - self.yl)

        return Profile(self.x, yu, self.x, yl)

    @property
    def tc(self):
        """thickness to chord ratio of the Profile"""
        return max(self.yu - self.yl)

    def set_tc(self, new_tc):
        factor = new_tc / self.tc

        self.yu *= factor
        self.yl *= factor



"""Provides the 'OffshoreSubstationDesign` class."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "Jake.Nunemaker@nrel.gov"


import numpy as np

from wisdem.orbit.phases.design import DesignPhase


class OffshoreSubstationDesign(DesignPhase):
    """Offshore Substation Design Class."""

    expected_config = {
        "site": {"depth": "m"},
        "plant": {"num_turbines": "int"},
        "turbine": {"turbine_rating": "MW"},
        "substation_design": {
            "mpt_cost_rate": "USD/MW (optional)",
            "topside_fab_cost_rate": "USD/t (optional)",
            "topside_design_cost": "USD (optional)",
            "shunt_cost_rate": "USD/MW (optional)",
            "switchgear_cost": "USD (optional)",
            "backup_gen_cost": "USD (optional)",
            "workspace_cost": "USD (optional)",
            "other_ancillary_cost": "USD (optional)",
            "topside_assembly_factor": "float (optional)",
            "oss_substructure_cost_rate": "USD/t (optional)",
            "oss_pile_cost_rate": "USD/t (optional)",
            "num_substations": "int (optional)",
        },
    }

    output_config = {
        "num_substations": "int",
        "offshore_substation_topside": "dict",
        "offshore_substation_substructure": "dict",
    }

    def __init__(self, config, **kwargs):
        """
        Creates an instance of OffshoreSubstationDesign.

        Parameters
        ----------
        config : dict
        """

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)
        self._outputs = {}

    def run(self):
        """Main run function."""

        self.calc_substructure_length()
        self.calc_substructure_deck_space()
        self.calc_topside_deck_space()

        self.calc_num_mpt_and_rating()
        self.calc_mpt_cost()
        self.calc_topside_mass_and_cost()
        self.calc_shunt_reactor_cost()
        self.calc_switchgear_cost()
        self.calc_ancillary_system_cost()
        self.calc_assembly_cost()
        self.calc_substructure_mass_and_cost()

        self._outputs["offshore_substation_substructure"] = {
            "type": "Monopile",  # Substation install only supports monopiles
            "deck_space": self.substructure_deck_space,
            "mass": self.substructure_mass,
            "length": self.substructure_length,
            "unit_cost": self.substructure_cost,
        }

        self._outputs["offshore_substation_topside"] = {
            "deck_space": self.topside_deck_space,
            "mass": self.topside_mass,
            "unit_cost": self.substation_cost,
        }

        self._outputs["num_substations"] = self.num_substations

    @property
    def substation_cost(self):
        """Returns total procuremet cost of the topside."""

        return (
            self.mpt_cost
            + self.topside_cost
            + self.shunt_reactor_cost
            + self.switchgear_costs
            + self.ancillary_system_costs
            + self.land_assembly_cost
        )

    @property
    def total_cost(self):
        """Returns total procurement cost of the substation(s)."""

        if not self._outputs:
            raise Exception("Has OffshoreSubstationDesign been ran yet?")

        return (self.substructure_cost + self.substation_cost) * self.num_substations

    def calc_substructure_length(self):
        """
        Calculates substructure length as the site depth + 10m
        """

        self.substructure_length = self.config["site"]["depth"] + 10

    def calc_substructure_deck_space(self):
        """
        Calculates required deck space for the substation substructure.

        Coming soon!
        """

        self.substructure_deck_space = 1

    def calc_topside_deck_space(self):
        """
        Calculates required deck space for the substation topside.

        Coming soon!
        """

        self.topside_deck_space = 1

    def calc_num_mpt_and_rating(self):
        """
        Calculates the number of main power transformers (MPTs) and their rating.

        Parameters
        ----------
        num_turbines : int
        turbine_rating : float
        """

        _design = self.config.get("substation_design", {})

        num_turbines = self.config["plant"]["num_turbines"]
        turbine_rating = self.config["turbine"]["turbine_rating"]
        capacity = num_turbines * turbine_rating

        self.num_substations = _design.get("num_substations", int(np.ceil(capacity / 500)))
        self.num_mpt = np.ceil(num_turbines * turbine_rating / (250 * self.num_substations))
        self.mpt_rating = (
            round(((num_turbines * turbine_rating * 1.15) / (self.num_mpt * self.num_substations)) / 10.0) * 10.0
        )

    def calc_mpt_cost(self):
        """
        Calculates the total cost for all MPTs.

        Parameters
        ----------
        mpt_cost_rate : float
        """

        _design = self.config.get("substation_design", {})
        mpt_cost_rate = _design.get("mpt_cost_rate", 12500)

        self.mpt_cost = self.mpt_rating * self.num_mpt * mpt_cost_rate

    def calc_topside_mass_and_cost(self):
        """
        Calculates the mass and cost of the substation topsides.

        Parameters
        ----------
        topside_fab_cost_rate : int | float
        topside_design_cost: int | float
        """

        _design = self.config.get("substation_design", {})
        topside_fab_cost_rate = _design.get("topside_fab_cost_rate", 14500)
        topside_design_cost = _design.get("topside_design_cost", 4.5e6)

        self.topside_mass = 3.85 * self.mpt_rating * self.num_mpt + 285
        self.topside_cost = self.topside_mass * topside_fab_cost_rate + topside_design_cost

    def calc_shunt_reactor_cost(self):
        """
        Calculates the cost of the shunt reactor.

        Parameters
        ----------
        shunt_cost_rate : int | float
        """

        _design = self.config.get("substation_design", {})
        shunt_cost_rate = _design.get("shunt_cost_rate", 35000)

        self.shunt_reactor_cost = self.mpt_rating * self.num_mpt * shunt_cost_rate * 0.5

    def calc_switchgear_cost(self):
        """
        Calculates the cost of the switchgear.

        Parameters
        ----------
        switchgear_cost : int | float
        """

        _design = self.config.get("substation_design", {})
        switchgear_cost = _design.get("switchgear_cost", 14.5e5)

        self.switchgear_costs = self.num_mpt * switchgear_cost

    def calc_ancillary_system_cost(self):
        """
        Calculates cost of ancillary systems.

        Parameters
        ----------
        backup_gen_cost : int | float
        workspace_cost : int | float
        other_ancillary_cost : int | float
        """

        _design = self.config.get("substation_design", {})
        backup_gen_cost = _design.get("backup_gen_cost", 1e6)
        workspace_cost = _design.get("workspace_cost", 2e6)
        other_ancillary_cost = _design.get("other_ancillary_cost", 3e6)

        self.ancillary_system_costs = backup_gen_cost + workspace_cost + other_ancillary_cost

    def calc_assembly_cost(self):
        """
        Calculates the cost of assembly on land.

        Parameters
        ----------
        topside_assembly_factor : int | float
        """

        _design = self.config.get("substation_design", {})
        topside_assembly_factor = _design.get("topside_assembly_factor", 0.075)
        self.land_assembly_cost = (
            self.switchgear_costs + self.shunt_reactor_cost + self.ancillary_system_costs
        ) * topside_assembly_factor

    def calc_substructure_mass_and_cost(self):
        """
        Calculates the mass and associated cost of the substation substructure.

        Parameters
        ----------
        oss_substructure_cost_rate : int | float
        oss_pile_cost_rate : int | float
        """

        _design = self.config.get("substation_design", {})
        oss_substructure_cost_rate = _design.get("oss_substructure_cost_rate", 3000)
        oss_pile_cost_rate = _design.get("oss_pile_cost_rate", 0)

        substructure_mass = 0.4 * self.topside_mass
        substructure_pile_mass = 8 * substructure_mass**0.5574
        self.substructure_cost = (
            substructure_mass * oss_substructure_cost_rate + substructure_pile_mass * oss_pile_cost_rate
        )

        self.substructure_mass = substructure_mass + substructure_pile_mass

    @property
    def design_result(self):
        """
        Returns the results of self.run().
        """

        if not self._outputs:
            raise Exception("Has OffshoreSubstationDesign been ran yet?")

        return self._outputs

    @property
    def detailed_output(self):
        """Returns detailed phase information."""

        _outputs = {
            "num_substations": self.num_substations,
            "substation_mpt_rating": self.mpt_rating,
            "substation_topside_mass": self.topside_mass,
            "substation_topside_cost": self.topside_cost,
            "substation_substructure_mass": self.substructure_mass,
            "substation_substructure_cost": self.substructure_cost,
        }

        return _outputs



"""
John Jasa
May 2020

This script parses a file containing OpenMDAO components and produces
docstrings for those components. It makes some assumptions about the
format of the code. Specifically, it reads the `add_input` and `add_output`
statements from the components and generates docstrings based on that.

Example usage:
python convert_docstrings.py tower.py

This creates a file called `tower_docstrings.py` which contains the
modified components. These new components can then be used as-is or copied
and pasted into the older file.
"""

import sys
import textwrap

filename = sys.argv[1]
with open(filename) as f:
    content = f.readlines()
orig_content = content.copy()
orig_content = [x.strip("\n") for x in orig_content]
content = [x.strip() for x in content]

doc_lines = []

in_class_flag = False
for line in content:
    if len(line) > 0:
        if "#" in line[0]:
            continue

    if "Component):" in line:
        class_name = line.split()[1]
        in_class_flag = True

        parsed_data = {}
        parsed_data["inputs"] = {}
        parsed_data["outputs"] = {}
        new_add_statements = []

    elif (
        ".add_input(" in line
        or ".add_output(" in line
        or ".add_discrete_input(" in line
        or ".add_discrete_output(" in line
    ):
        if in_class_flag:
            if "_input(" in line:
                is_input = True
            else:
                is_input = False

            if "_discrete_" in line:
                discrete_string = "discrete_"
            else:
                discrete_string = ""

            # Split statements on commas to roughly group the terms
            comma_sep_line = line.split(",")

            # Strip whitespace from each entry
            stripped_line = [x.strip() for x in comma_sep_line]

            # Need this logic to handle if the name used '' or ""
            if is_input:
                var_name = stripped_line[0].split(f"add_{discrete_string}input(")[1][1:-1]
            else:
                var_name = stripped_line[0].split(f"add_{discrete_string}output(")[1][1:-1]

            # Get the value type from the line
            val_str = stripped_line[1]

            # Some logic to handle numpy arrays
            left = val_str.count("(")
            right = val_str.count(")")
            i_paren = 1
            while left != right:
                val_str += ", " + stripped_line[1 + i_paren]
                i_paren += 1
                right = val_str.count(")")
            if "=" in val_str:
                val_str = val_str.split("=")[1].strip()

            if "np" in val_str:
                split_str = val_str.split("(")[-1].split(")")[0]
                type_ = f"numpy array[{split_str}]"
                type_ = type_.replace("[[", "[")
                type_ = type_.replace("]]", "]")
            else:
                try:
                    float(val_str)
                    if "discrete" in discrete_string:
                        type_ = "TODO: add type by hand, could not be parsed automatically"
                    else:
                        type_ = "float"

                except:
                    type_ = "TODO: add type by hand, could not be parsed automatically"

            if "desc" in line:
                desc = line.split("desc")[1]
                if "'" in desc:
                    desc = desc.split("'")[1]
                elif '"' in desc:
                    desc = desc.split('"')[1]

                desc = " ".join([x.strip() for x in desc.split()])
                if len(desc) > 80:
                    desc = "\n        ".join(textwrap.wrap(desc, 80))

                split_line = line.split(",")
                new_split_line = []
                for i_string, substring in enumerate(split_line):
                    if "desc" in substring:
                        if i_string == len(split_line) - 1:
                            desc_last = True
                        else:
                            desc_last = False
                    else:
                        new_split_line.append(substring)

                # Need special logic is desc is the last entry in the add_ call
                # or if units comes after
                if desc_last:
                    new_add_statement = line.split("desc")[0].strip()[:-1] + ")"
                else:
                    new_add_statement = ", ".join(new_split_line)

                new_add_statements.append(new_add_statement)

            else:
                desc = ""
                new_add_statements.append(line)

            # If units are included, parse the line to get the actual units
            if "units" in line:
                # Split the line on the word units; grab everything after that
                post_units = line.split("units")[1]

                # Remove all whitespace from the string
                post_units = "".join(post_units.split())

                # Split the line on ',' or ')' to remove extra formatting
                post_units = post_units.split(",")[0]
                post_units = post_units.split(")")[0]

                # Get only the actual units, no quotation marks or equals signs
                units = post_units[2:-1]

                # Add formatting to make it look good on the docstring
                units = f", [{units}]"

            # If no units in line, add placeholder string
            else:
                units = ""

            if is_input:
                parsed_data["inputs"][var_name] = {"type_": type_, "desc": desc, "units": units}
            else:
                parsed_data["outputs"][var_name] = {"type_": type_, "desc": desc, "units": units}

    elif "compute(" in line:
        in_class_flag = False

        doc_lines.append(class_name)
        doc_lines.append('"""')
        doc_lines.append("TODO : replace this with docstring")
        doc_lines.append("")

        doc_lines.append("Parameters")
        doc_lines.append("----------")
        for var_name in parsed_data["inputs"]:
            data = parsed_data["inputs"][var_name]
            doc_lines.append(f"{var_name} : {data['type_']}{data['units']}")
            doc_lines.append(f"    {data['desc']}")

        doc_lines.append("")
        doc_lines.append("Returns")
        doc_lines.append("-------")
        for var_name in parsed_data["outputs"]:
            data = parsed_data["outputs"][var_name]
            doc_lines.append(f"{var_name} : {data['type_']}{data['units']}")
            doc_lines.append(f"    {data['desc']}")
        doc_lines.append("")
        doc_lines.append('"""')

        doc_lines.append("")
        doc_lines.append("")
        output_flag = False
        for add_line in new_add_statements:
            if "_output" in add_line and not output_flag:
                doc_lines.append("")
                output_flag = True
            stripped_line = ", ".join([x.strip() for x in add_line.split(",")])
            doc_lines.append(stripped_line)

        doc_lines.append("")
        doc_lines.append("")

for i_line, line in enumerate(doc_lines):
    doc_lines[i_line] = f"    {line}"

new_filename = filename.split(".")[0] + "_docstrings." + filename.split(".")[1]

full_file = []
in_class_flag = False
i_docs = 0
for line in orig_content:
    if len(line) > 0:
        if "#" in line[0]:
            full_file.append(line)
            continue

    if "Component):" in line:
        class_name = line.split()[1]
        in_class_flag = True
        full_file.append(line)
        adds_done = False

        smaller_doc_lines = doc_lines[i_docs:]
        in_docstring = False
        for new_line in smaller_doc_lines:
            if '"""' in new_line:
                if not in_docstring:
                    in_docstring = True
                else:
                    in_docstring = False
                    full_file.append(new_line)
                    i_docs += 2
                    break
            if in_docstring:
                full_file.append(new_line)
                i_docs += 1

    elif "self.add_" in line and in_class_flag and not adds_done:
        smaller_doc_lines = doc_lines[i_docs:]
        in_adds = False
        first_output = True
        for new_line in smaller_doc_lines:
            if "self.add_" in new_line:
                if "_output" in new_line and first_output:
                    full_file.append("")
                    first_output = False

                full_file.append(f"    {new_line}")
                i_docs += 1
            elif "Component)" in new_line:
                i_docs += 1
                adds_done = True
                break
            else:
                if len(new_line.strip()) > 0:
                    full_file.append(f"    {new_line}")
                i_docs += 1

    elif "compute(" in line:
        in_class_flag = False
        full_file.append(line)

    elif "self.add_" in line and "add_subsystem" not in line:
        pass

    else:
        full_file.append(line)

with open(new_filename, "w") as f:
    for line in full_file:
        f.write(line + "\n")



import traceback
from wisdem.landbosse.model.CostModule import CostModule
import pandas as pd
import math

class DevelopmentCost(CostModule):
    """
    **DevelopmentCost.py

    -Created by Parangat Bhaskar on June 30, 2019

    Creating a simple DevelopmentCost module for now. This module reads in a user input from the detailed input Excel
    file and outputs this in the detailed output excel file.

    """

    def __init__(self, input_dict, output_dict, project_name):
        self.input_dict = input_dict
        self.output_dict = output_dict
        self.project_name = project_name

    def calculate_costs(self):
        """
        Sets the total cost for development.

        If there is a key of 'development_labor_cost_usd' in the input
        dictionary, then that is used as the development cost and a
        dataframe is created that holds that labor cost.

        If the key is not present in the dictionary, the development
        cost is retrieved from the project data.

        Returns
        ------------------------------------
        total_development_cost : pd.DataFrame
            data frame with total development cost by type of cost (e.g., Labor)
        """
        if 'development_labor_cost_usd' in self.input_dict:
            total_development_cost = pd.DataFrame([
                {'Type of cost': 'Equipment rental', 'Cost USD': 0, 'Phase of construction': 'Development'},
                {'Type of cost': 'Labor', 'Cost USD': self.input_dict['development_labor_cost_usd'],
                 'Phase of construction': 'Development'},
                {'Type of cost': 'Materials', 'Cost USD': 0, 'Phase of construction': 'Development'},
                {'Type of cost': 'Mobilization', 'Cost USD': 0, 'Phase of construction': 'Development'},
                {'Type of cost': 'Other', 'Cost USD': 0, 'Phase of construction': 'Development'}
            ])
        else:
            total_development_cost = self.input_dict['development_df']

        self.output_dict['total_development_cost'] = total_development_cost

        return total_development_cost

    def outputs_for_detailed_tab(self):
        """
        Creates a list of dictionaries which can be used on their own or
        used to make a dataframe.

        Must be called after self.run_module()

        Returns
        -------
        list(dict)
            A list of dicts, with each dict representing a row of the data.
        """

        result = []
        module = type(self).__name__
        for _, row in self.output_dict['total_development_cost'].iterrows():
            dashed_row = '{} - {} - {}'.format(row["Type of cost"], row["Phase of construction"], math.ceil(row["Cost USD"]))
            result.append({
                'unit': '',
                'type': 'dataframe',
                'variable_df_key_col_name': 'Type of Cost - Phase of Construction - Cost in USD',
                'value': dashed_row,
                'last_number': row.iloc[2]
            })

        for _dict in result:
            _dict['project_id_with_serial'] = self.project_name
            _dict['module'] = module

        self.output_dict['development_cost_csv'] = result
        return result

    def run_module(self):
        """
        Runs the DevelopmentCost module and populates the IO dictionaries with calculated values.

        """

        try:
            self.calculate_costs()
            self.outputs_for_detailed_tab()
            self.output_dict['development_module_type_operation'] = self.outputs_for_costs_by_module_type_operation(
                input_df=self.output_dict['total_development_cost'],
                project_id=self.project_name,
                total_or_turbine=True
            )
            return 0, 0  # module ran successfully

        except Exception as error:
            traceback.print_exc()
            print(f"Fail {self.project_name} DevelopmentCost")
            return 1, error  # module did not run successfully





"""Provides the `ArraySystemDesign` class."""

__author__ = "Rob Hammond"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Rob Hammond"
__email__ = "robert.hammond@nrel.gov"


import warnings
from collections import OrderedDict

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from wisdem.orbit.core.library import export_library_specs, extract_library_specs
from wisdem.orbit.phases.design._cables import Plant, CableSystem


class ArraySystemDesign(CableSystem):
    """
    The design phase for an array cabling system.

    Attributes
    ----------
    system : `Plant`
        A `Plant` object for wind farm specifications.
    cables : `collections.OrderedDict`
        A dictionary of `Cable` objects sorted from smallest to largest.
    full_string : list
        A list of `Cable.name` that represents the longest (in turbine
        connections) possible string.
    partial_string : list
        A list of `Cable.name` that represents the leftover turbines that could
        not create another `full_string` to complete the layout.
    num_turbines_full_string : int
        Number of turbines on a `full_string`.
    num_turbines_partial_string : int
        Number of turbines on a `partial_string`.
    num_full_strings : int
        Number of full strings required to connect the wind farm.
    num_partial_strings : int
        Number of partial strings required to connect the wind farm.
    num_strings : int
        Total number of strings in the system.
    turbines_x : np.ndarray, [`num_strings`, `num_turbines_full_string`]
        The relative x-coordinates of every turbine.
    turbines_y : np.ndarray, [`num_strings`, `num_turbines_full_string`]
        The relative y-coordinates of every turbines.
    oss_x : float
        The relative x-coordinates for the OSS.
    oss_y : float
        The relative y-coordinates for the OSS.
    coordinates : np.ndarray, [`num_strings`, `num_turbines_full_string` + 1, 2]
        The relative (x, y) coordinates for the entire wind farm with the
        first column representing the OSS.
    sections_distances : np.ndarray, [`num_strings`, `num_turbines_full_string`]
        The Euclidean distance between any two points in a string extending
        outward from the OSS.
    sections_cable_lengths : np.ndarray, [`num_strings`, `num_turbines_full_string`]
        `sections_distances` + 2 * `system.site_depth` to account for the water
        depth at each turbine.
    sections_cables : np.ndarray, [`num_strings`, `num_turbines_full_string`]
        The type of cable being used to connect turbines in a string. All
        values are either ``None`` or `Cable.name`.
    """

    expected_config = {
        "site": {"depth": "m"},
        "plant": {
            "layout": "str",
            "row_spacing": "rotor diameters",
            "turbine_spacing": "rotor diameters",
            "turbine_distance": "km (optional)",
            "num_turbines": "int",
            "substation_distance": "km",
        },
        "turbine": {"rotor_diameter": "m", "turbine_rating": "MW"},
        "array_system_design": {
            "cables": "list | str",
            "touchdown_distance": "m (optional, default: 0)",
            "average_exclusion_percent": "float (optional)",
            "floating_cable_depth": "m (optional, default: water depth)",
        },
    }

    output_config = {"array_system": {"cables": "dict", "system_cost": "USD"}}

    def __init__(self, config, **kwargs):
        """
        Defines the cables and sections required to install an offshore wind
        farm.

        Parameters
        ----------
        config : dict
            Configuration dictionary for the array cabling system. See
            `expected_config` for details on what is required.
        """

        super().__init__(config, "array", **kwargs)

        self.exclusion = 1 + self.config["array_system_design"].get("average_exclusion_percent", 0.0)
        self._get_touchdown_distance()
        self.extract_phase_kwargs(**kwargs)
        self.system = Plant(self.config)

    @property
    def total_length(self):
        """Returns total array system length."""

        return sum([v for _, v in self.total_cable_length_by_type.items()])

    @property
    def total_cable_cost(self):
        """Returns total array system cable cost."""

        return sum(self.cost_by_type.values())

    @property
    def detailed_output(self):
        """Returns array system design outputs."""

        _output = {
            **self.design_result,
            "array_system_num_strings": self.num_strings,
            "array_system_total_length": self.total_length,
            "array_system_length_by_type": self.total_cable_length_by_type,
            "array_system_total_cost": self.total_cable_cost,
            "array_system_cost_by_type": self.cost_by_type,
            "array_system_num_turbines_full_string": self.num_turbines_full_string,
            "array_system_num_full_strings": self.num_full_strings,
        }

        return _output

    def _compute_euclidean_distance(self):
        """Calculates the distance between two cartesian coordinate points.

        Returns
        -------
        np.ndarray
            The Euclidean distance between subsequent pairs of turbines in a
            string for all strings in the windfarm.
        """
        differnce = np.abs(np.diff(self.coordinates, n=1, axis=1))
        distance = np.round_(np.linalg.norm(differnce, axis=2), 10)
        return distance

    def _compute_maximum_turbines_per_cable(self):
        """
        Calculates the maximum turbines that each cable can support and adds
        it to the `Cable` object.
        """

        for cable in self.cables.values():
            with np.errstate(divide="ignore", invalid="ignore"):
                cable.max_turbines = np.floor(cable.cable_power / self.system.turbine_rating)
                if cable.max_turbines == float("inf"):
                    raise ValueError("Must be at least 1 turbine in windfarm!")

    def _compute_string(self, max_turbines_per_string):
        """
        Calculates the maximum number of turbines that each cable type
        can support and builds a string (as a list) of cable types that
        is of length of `max_turbines_per_string` or the maximum number
        of turbines that the largest cable can support, whichever is lower.

        Parameters
        ----------
        max_turbines_per_string : int
            The maximum number of turbines that can be supported on a single
            string given the cable(s) properties.

        Returns
        -------
        cable_layout : list
            A list of cables required to create an the maximum length string
            possible given `cables` and `max_turbines_per_string`. The order
            of the cabling is from largest to smallest cable type with the
            `Cable.name` as the list entry.
        """

        cable_layout = []
        _cables = list(self.cables.values())

        n = len(cable_layout)
        while n < max_turbines_per_string and _cables:
            _cable = _cables.pop(0)

            # Ensure that the most turbines in a string is is lower than the
            # string maximum and the maximum the individual cable can support,
            # then add another cable.
            while max_turbines_per_string > n < _cable.max_turbines:
                cable_layout.append(_cable.name)
                n = len(cable_layout)

        # Return the reverse to start from the largest cable to the smallest
        # cable to represent the string extending from the OSS.
        return cable_layout[::-1]

    def create_strings(self):
        """
        Calculates the required full and partial string design.

        .. note:: For custom layouts this is to provide guidance on the number
        of strings.
        """

        self._compute_maximum_turbines_per_cable()

        # Create the longest string possible with the given cable properties
        self.full_string = self._compute_string(list(self.cables.values())[-1].max_turbines)
        self.num_turbines_full_string = len(self.full_string)
        self.num_full_strings, self.num_turbines_partial_string = np.divmod(
            self.system.num_turbines, self.num_turbines_full_string
        )

        # Create a partial string constrained by the remainder
        self.partial_string = self._compute_string(self.num_turbines_partial_string)
        self.num_partial_strings = 1 if self.num_turbines_partial_string > 0 else 0

        self.num_strings = self.num_full_strings + self.num_partial_strings

    def _design_grid_layout(self):
        """
        Makes the coordinates of a default grid layout.
        """

        # Create the relative (x, y) coordinate matrices for the turbines
        # using vector math

        # X = column vector of turbine distance
        #     * row vector of range(1, num_turbines_full_string + 1)
        self.turbines_x = np.full(self.num_strings, self.system.turbine_distance).reshape(-1, 1) * np.add(
            np.arange(self.num_turbines_full_string, dtype=float), 1
        )

        # Y = column vector of reverse range(1, num_strings)
        #     * row vector of row_distance
        self.turbines_y = np.arange(self.num_strings, dtype=float)[::-1].reshape(-1, 1) * np.full(
            (1, self.num_turbines_full_string), self.system.row_distance
        )

        # If there are partial strings the default layout then null out
        # the non-existent turbines
        if self.partial_string:
            self.turbines_x[-1, self.num_turbines_partial_string :] = None
            self.turbines_y[-1, self.num_turbines_partial_string :] = None

        # Create the relative OSS coordinates
        self.oss_x = 0.0
        self.oss_y = self.turbines_y[:, 0].mean()

    def _design_ring_layout(self):
        """
        Creates the coordinates of a default ring layout.
        """

        # Calculate the radius of each turbine from the OSS
        radius = (
            np.arange(self.num_turbines_full_string, dtype=float) * self.system.turbine_distance
            + self.system.substation_distance
        )

        # Calculate the angle required to make rings evenly spaced
        radians = np.linspace(0, 2 * np.pi, self.num_strings + 1)

        # Calculate the relative x and y coordinates for the turbines
        x = np.sin(radians).reshape(-1, 1) * radius
        y = np.cos(radians).reshape(-1, 1) * radius

        # Remove the repeated row that completes the circle
        self.turbines_x = np.delete(x, (-1), axis=0)
        self.turbines_y = np.delete(y, (-1), axis=0)

        # Remove the extra turbines from the partial string if there is one
        if self.partial_string:
            self.turbines_x[-1, self.num_turbines_partial_string :] = None
            self.turbines_y[-1, self.num_turbines_partial_string :] = None

        # Create the relative OSS coordinates
        self.oss_x = 0.0
        self.oss_y = 0.0

    def _create_wind_farm_layout(self):
        """
        Creates a list of substation-to-string distances based off the layout
        of the wind farm.
        """

        if self.system.layout == "custom":
            raise NotImplementedError("Use `CustomArraySystemDesign` for custom layouts")

        if self.system.layout == "grid":
            self._design_grid_layout()

        if self.system.layout == "ring":
            self._design_ring_layout()

        # Create the relative wind farm coordinates
        self.coordinates = np.dstack(
            (
                np.insert(self.turbines_x, 0, self.oss_x, axis=1),
                np.insert(self.turbines_y, 0, self.oss_y, axis=1),
            )
        )

        # Take the norm of the difference of turbine "coordinate pairs".
        self.sections_distance = self._compute_euclidean_distance()

    def _create_cable_section_lengths(self):
        """
        For each cable compute the number of section lengths required to
        complete the array cabling.
        """

        if getattr(self, "sections_cable_lengths", np.zeros(1)).sum() == 0:
            self.sections_cable_lengths = (
                self.sections_distance * self.exclusion + (2 * self.free_cable_length) - (2 * self.touchdown / 1000)
            )
        self.sections_cables = np.full((self.num_strings, self.num_turbines_full_string), None)

        # Create an array of the cable names for each cable section
        for i, row in enumerate(np.isnan(self.coordinates[:, :, 0])):
            ix = row.sum()
            # Fill each string (row in array) with the cables names for valid
            # positions in the layout starting from the end
            self.sections_cables[i, 0 : self.num_turbines_full_string - ix] = self.full_string[::-1][
                : self.num_turbines_full_string - ix
            ][::-1]

    def run(self):
        """
        Runs all the functions to create an array sytem.
        """

        self._initialize_cables()
        self.create_strings()
        self._create_wind_farm_layout()
        self._create_cable_section_lengths()

    def save_layout(self, save_name, return_df=False):
        """Outputs a csv of the substation and turbine positional and cable
        related components.

        Parameters
        ----------
        save_name : str
            The name of the file without an extension to be saved to
            <library_path>/cables/<save_name>.csv.
        return_df : bool, optional
            If true, returns layout_df, a pandas.DataFrame of the cabling
            layout, by default False.

        Returns
        -------
        pd.DataFrame
            The DataFrame with the layout data.
        """
        num_turbines = self.system.num_turbines
        columns = [
            "id",
            "substation_id",
            "name",
            "latitude",
            "longitude",
            "string",
            "order",
            "cable_type",
            "euclidean_distance",
            "cable_length",
        ]
        layout_df = pd.DataFrame(
            np.zeros((self.system.num_turbines + 1, len(columns))),
            columns=columns,
        )
        layout_df.string = layout_df.string.astype(int)
        layout_df.order = layout_df.order.astype(int)

        strings = [("", "")]
        strings.extend([(i, j) for i in range(self.num_full_strings) for j in range(self.num_turbines_full_string)])
        try:
            strings.extend(
                [
                    (i, j)
                    for i in range(self.num_full_strings, self.num_strings)
                    for j in range(self.num_turbines_partial_string)
                ]
            )
        except AttributeError:
            pass
        layout_df[["string", "order"]] = strings

        coords = np.array([[0, 0]] + list(zip(self.turbines_x.flatten(), self.turbines_y.flatten())))
        coords = coords[: self.system.num_turbines + 1]
        layout_df[["longitude", "latitude"]] = coords

        layout_df["substation_id"] = "oss1"
        layout_df["id"] = ["oss1"] + [f"t{i}" for i in range(layout_df.shape[0] - 1)]
        layout_df["name"] = ["substation-1"] + [f"turbine-{i}" for i in range(num_turbines)]
        layout_df.cable_type = [""] + self.sections_cables.flatten()[:num_turbines].tolist()
        layout_df.euclidean_distance = [""] + self.sections_distance.flatten()[:num_turbines].tolist()
        layout_df.cable_length = [""] + self.sections_cable_lengths.flatten()[:num_turbines].tolist()
        data = [columns] + layout_df.values.tolist()
        print(f"Saving custom array CSV to: <library_path>/cables/{save_name}.csv")
        export_library_specs("cables", save_name, data, file_ext="csv")
        if return_df:
            return layout_df

    def _plot_oss(self, ax):
        """
        Adds the offshore substation(s) to the plot.

        Colors are a selection of the colorblind pallette from Bokeh:
        https://github.com/bokeh/bokeh/blob/master/bokeh/palettes.py#L1938-L1940

        Parameters
        ----------
        ax : matplotlib.axes
            Axis object to add the substation(s) to.
        """

        #        ['orange', 'blugren', 'skyblue', 'vermill', 'redprpl']
        colors = ["#E69F00", "#009E73", "#56B4E9", "#D55E00", "#CC79A7"]
        kwargs = {"s": 300, "c": "#E69F00", "zorder": 2}

        # If this is a custom layout there could be multiple OSS to consider
        if getattr(self, "windfarm", None) is not None:
            labels_set = []
            locations = self.location_data.drop_duplicates(
                subset=[
                    "substation_id",
                    "substation_name",
                    "substation_longitude",
                    "substation_latitude",
                ],
                keep="first",
            )
            for ix, (_, row) in enumerate(locations.iterrows()):
                name = row.substation_name
                name = row.substation_id if name == 0 else name
                labels_set.append(name)
                kwargs["c"] = colors[ix]
                ax.scatter(
                    row.substation_longitude,
                    row.substation_latitude,
                    label=name,
                    **kwargs,
                )
        else:
            labels_set = ["OSS"]
            ax.scatter(self.oss_x, self.oss_y, label="OSS", **kwargs)

        return labels_set + ["Turbine"], ax

    def plot_array_system(self, show=True, save_path_name=None, return_fig=False):
        """
        Plot the array cabling system.

        Parameters
        ----------
        show : bool, default: True
            If True the plot will be output inline or to screen.
        save_path_name : str, default: None
            The <path_to_file>/<file_name> of where to save the created plot.
            If None then the plot will not be saved.
        return_fig : bool, default: False
            If true, the figure (`fig`) and axes (`ax`) objects will be
            returned.

        Returns
        -------
        fig : matplotlib.pyplot.figure
            The matplotlib figure object. This will contain the overarching
            figure settings, and can be manipulated to change dimensions,
            resolotion, and etc.
        ax : matplotlib.pyplot.axes
            The matplotlib axes object. This will contain the actual plot
            settings, and can be manipulated to add annotations, or other
            elements.
        """

        fig, ax = plt.subplots(figsize=(10, 10))
        plt.axis("off")

        # Plot the offshore substation and turbiness
        labels_set, ax = self._plot_oss(ax)
        ax.scatter(
            self.coordinates[:, 1:, 0],
            self.coordinates[:, 1:, 1],
            marker="1",
            s=250,
            c="#0072B2",
            linewidth=2,
            zorder=2,
            label="Turbine",
        )
        # Plot the turbine names
        # for i in range(self.coordinates.shape[0]):
        #     for j in range(self.coordinates.shape[1] - 1):
        #         if not np.any(np.isnan(self.coordinates[i, j + 1])):
        #             x, y = self.coordinates[i, j + 1]
        #             name = self.location_data.loc[(self.location_data.string == i) & (self.location_data.order == j), "turbine_name"].values[0]
        #             ax.text(x, y, name)

        # Determine the cable section widths
        string_sets = np.unique(
            [list(OrderedDict.fromkeys(el for el in cables if el is not None)) for cables in self.sections_cables]
        )
        if isinstance(string_sets[0], list):
            max_string = max(string_sets, key=len)
        else:
            max_string = string_sets[::-1]
        string_widths = np.arange(len(max_string), 0, -1, dtype=float).tolist()

        for i, row in enumerate(self.sections_cables):
            for cable, width in zip(max_string, string_widths):
                ix_to_plot = np.where(row == cable)[0]
                if ix_to_plot.size == 0:
                    continue
                ix_to_plot = np.append(ix_to_plot, ix_to_plot.max() + 1)

                ax.plot(
                    self.coordinates[i, ix_to_plot, 0],
                    self.coordinates[i, ix_to_plot, 1],
                    linewidth=width,
                    color="k",
                    zorder=1,
                    label=cable,
                )

        # Make the plot range larger to evenly contain the legend
        lim_range = np.diff(ax.get_xlim())[0] / 20
        shift = lim_range * np.array([-1, 1])
        ax.set_xlim(np.array(ax.get_xlim()) + shift)
        ax.set_ylim(np.array(ax.get_ylim()) + shift)

        # Constrain the legend to only have on of each type of cable
        handles, labels = ax.get_legend_handles_labels()
        labels_set.extend([str(el) for el in max_string])
        ix_filter = [labels.index(el) for el in labels_set]
        handles = [handles[ix] for ix in ix_filter]
        labels = [labels[ix] for ix in ix_filter]
        ax.legend(
            handles,
            labels,
            ncol=(2 + len(max_string)),
            loc="lower center",
            mode="expand",
            borderpad=0.7,
            borderaxespad=-0.4,
            labelspacing=1.2,
        )

        plt.tight_layout()
        if save_path_name is not None:
            plt.savefig(save_path_name, bbox_inches="tight", dpi=360)
        if show:
            plt.show()
        if return_fig:
            return fig, ax


class CustomArraySystemDesign(ArraySystemDesign):
    """
    Custom array system design phase.

    Parameters
    ----------
    ArraySystemDesign : ArraySystemDesign
        Array system design phase.

    Returns
    -------
    [type]
        [description]

    Raises
    ------
    ValueError
        [description]
    ValueError
        [description]
    ValueError
        [description]
    """

    expected_config = {
        "site": {"depth": "str"},
        "plant": {"layout": "str", "num_turbines": "int"},
        "turbine": {"turbine_rating": "int | float"},
        "array_system_design": {
            "cables": "list | str",
            "location_data": "str",
            "distance": "bool (optional)",
            "average_exclusion_percent": "float (optional)",
        },
    }

    # Columns that should be included in csv file.
    COLUMNS = [
        "id",
        "substation_id",
        "name",
        "latitude",
        "longitude",
        "string",
        "order",
        "cable_length",
        "bury_speed",
    ]

    # Transformed data column names
    REQUIRED = [
        "id",
        "substation_id",
        "substation_name",
        "substation_latitude",
        "substation_longitude",
        "turbine_name",
        "turbine_latitude",
        "turbine_longitude",
        "string",
        "order",
    ]
    OPTIONAL = ["cable_length", "bury_speed"]

    def __init__(self, config, distance=False, **kwargs):
        """
        Initializes the configuration.

        The location data must have the following columns:
        - substation_id : int
            ID for the substation.
        - substation_latitude : float
            Y-coordinate for a turbine's corresponding substation.
        - substation_longitude : float
            X-coordinate for a turbine's corresponding substation.
        - turbine_name : str
            Identifying name/ID for a turbine.
        - turbine_latitude : float
            Y-coordinate for the turbine.
        - turbine_longitude : float
            X-coordinate for the turbine.
        - string : int
            String number, starting from 0, the turbine is connected with.
        - order : int
            Turbine position on a string, starting from the OSS, starting
            from 0.

        Optional columns:
        - substation_name : str
            Descriptive name of the substation, default `substation_id`
        - cable_length : float
            Custom cable section length to account for exclusions that are not
            able to be accounted for with `average_exclusion_percent`,
            default 0.0, in km.
        - bury_speed : float
            Custom cable burying speeds for each cable section for highly
            variable soil conditions, by default 0.0 (uses vessel properites).

        Parameters
        ----------
        config : dict
            Configuration dictionary. See `expected_config`.
        distance : bool
            Indicator for reference coordinates, default False.
            - True: WGS84 latitude, longitude pairs for each coordinate
            - False: distance based pairs, in km.
        """

        super().__init__(config, **kwargs)
        self.distance = config["array_system_design"].get("distance", distance)

    def create_project_csv(self, save_name):
        """Creates a base CSV in <`library_path`>/cables/

        Parameters
        ----------
        save_name : [type]
            [description]
        """

        self._initialize_cables()
        self.create_strings()

        # Print the required output.
        rows = [
            ("N turbines full string", self.num_turbines_full_string),
            ("N full strings", self.num_full_strings),
            ("N turbines partial string", self.num_turbines_partial_string),
            ("N partial strings", self.num_partial_strings),
        ]
        border = "|"
        title = "PROJECT SPECIFICATIONS".rjust(26)
        sep = f"+{'-' * 32}+"
        print(sep)
        sep = "+".join((sep[:28], sep[29:]))
        print(f"{border} {title.ljust(30)} {border}")
        print(sep)
        for desc, n in rows:
            print(" ".join((border, desc.ljust(25), border, str(n).rjust(2), border)))
        print(sep)

        # Create the string order/number information
        strings = [(i, j) for i in range(self.num_full_strings) for j in range(self.num_turbines_full_string)]
        strings.extend(
            [
                (i, j)
                for i in range(self.num_full_strings, self.num_strings)
                for j in range(self.num_turbines_partial_string)
            ]
        )
        rows = [
            [f"t{i}", "oss1", f"turbine-{i}", 0.0, 0.0, string, order, 0, 0]
            for i, (string, order) in enumerate(strings)
        ]
        first = [
            "oss1",
            "oss1",
            "offshore_substation",
            0.0,
            0.0,
            "",
            "",
            "",
            "",
        ]
        rows.insert(0, first)
        rows.insert(0, self.COLUMNS)
        print(f"Saving custom array CSV to: <library_path>/cables/{save_name}.csv")
        export_library_specs("cables", save_name, rows, file_ext="csv")

    def _format_windfarm_data(self):
        # Separate the OSS data where substaion_id is equal to id
        substation_filter = self.location_data.substation_id == self.location_data.id
        oss = self.location_data[substation_filter].copy()
        oss.rename(
            columns={
                "latitude": "substation_latitude",
                "longitude": "substation_longitude",
                "name": "substation_name",
            },
            inplace=True,
        )
        oss.substation_id = oss["id"]
        oss.drop(
            ["id", "string", "order", "cable_length", "bury_speed"],
            inplace=True,
            axis=1,
        )

        # Separate the turbine data
        turbines = self.location_data[~substation_filter].copy()
        turbines.rename(
            columns={
                "latitude": "turbine_latitude",
                "longitude": "turbine_longitude",
                "name": "turbine_name",
            },
            inplace=True,
        )

        # Merge them back together
        self.location_data = turbines.merge(oss, on="substation_id", how="left")

        self.location_data = self.location_data[self.REQUIRED + self.OPTIONAL]

        self.location_data.string = self.location_data.string.astype(int)
        self.location_data.order = self.location_data.order.astype(int)
        self.location_data.sort_values(by=["substation_id", "string", "order"], inplace=True)

    def _initialize_custom_data(self):
        windfarm = self.config["array_system_design"]["location_data"]

        self.location_data = extract_library_specs("cables", windfarm, file_type="csv")

        # Make sure no data is missing
        missing = set(self.COLUMNS).difference(self.location_data.columns)
        if missing:
            raise ValueError(
                "The following columns must be included in the location ",
                f"data: {missing}",
            )

        self._format_windfarm_data()

        # Ensure there is no missing data in required columns
        missing_data_cols = [c for c in self.REQUIRED if pd.isnull(self.location_data[c]).sum() > 0]
        if missing_data_cols:
            raise ValueError(f"Missing data in columns: {missing_data_cols}!")

        # Ensure there is no missing data in optional columns
        missing_data_cols = [
            c for c in self.OPTIONAL if (pd.isnull(self.location_data[c]) | self.location_data[c] == 0).sum() > 0
        ]
        if missing_data_cols:
            message = f"Missing data in columns {missing_data_cols}; " "all values will be calculated."
            warnings.warn(message)

        # Ensure the number of turbines matches what's expected
        if self.location_data.shape[0] != self.system.num_turbines:
            raise ValueError(
                "The provided number of turbines ",
                f"({self.location_data.shape[0]}) does not match the plant ",
                f"data ({self.system.num_turbines}).",
            )

        n_coords = self.location_data.groupby(["turbine_latitude", "turbine_longitude"]).ngroups
        duplicates = self.location_data.shape[0] - n_coords
        if duplicates > 0:
            raise ValueError(f"There are {duplicates} rows with duplicate coordinates.")

        # Ensure the number of turbines on a string is within the limits
        longest_string = self.location_data["order"].unique().size
        self.num_strings = self.location_data.groupby(["substation_id", "string"]).ngroups
        if longest_string > self.num_turbines_full_string:
            raise ValueError("Strings can't contain more than " f"{self.num_turbines_full_string} turbines.")
        else:
            self.num_turbines_full_string = longest_string
            del self.num_turbines_partial_string
            del self.num_partial_strings

    def _check_optional_input(self):
        """
        Ensures that the optionally input parameters have valid data and were
        all filled out.
        """
        if np.any(self.sections_cable_lengths == 0):
            self.sections_cable_lengths = np.zeros((self.num_strings, self.num_turbines_full_string), dtype=float)

        if np.any(self.sections_bury_speeds == 0):
            self.sections_bury_speeds = np.zeros((self.num_strings, self.num_turbines_full_string), dtype=float)

    def _compute_haversine_distance(self):
        """Computes the haversine distance between two subsequent pairs in a
        string for all strings.

        Returns
        -------
        np.ndarray
            Haversine distance between all coordinate pairs in a string
        """
        RADIUS = 6371  # Radius of Earth in kilometers (3956 miles)
        coordinates_radians = np.radians(self.coordinates)

        lon1 = coordinates_radians[:, :-1, 0]
        lon2 = coordinates_radians[:, 1:, 0]
        lat1 = coordinates_radians[:, :-1, 1]
        lat2 = coordinates_radians[:, 1:, 1]

        dlat = lat2 - lat1
        dlon = lon2 - lon1

        a = np.sin(dlat / 2) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2) ** 2
        c = 2 * np.arcsin(np.sqrt(a))
        return c * RADIUS

    def _create_windfarm_layout(self):
        """
        Creates the custom windfarm layout that includes
            `windfarm_x`: x-coordinates with a corresponding OSS in the first
                column. Shape: `n_strings` x (`num_turbines_full_string` + 1).
            `windfarm_y`: y-coordinates with a corresponding OSS in the first
                column. Shape: `n_strings` x (`num_turbines_full_string` + 1).
            `sections_cables_lenghts`: custom cable lengths provided as an
                optional column in the `location_data`.
                Shape: `n_strings` x `num_turbines_full_string`.
            `sections_bury_speeds`: custom cable bury speeds provided as an
                optional column in the `location_data`.
                Shape: `n_strings` x `num_turbines_full_string`.
        """

        self.location_data_x = np.zeros((self.num_strings, self.num_turbines_full_string + 1), dtype=float)
        self.location_data_y = np.zeros((self.num_strings, self.num_turbines_full_string + 1), dtype=float)
        self.sections_cable_lengths = np.zeros((self.num_strings, self.num_turbines_full_string), dtype=float)
        self.sections_bury_speeds = np.zeros((self.num_strings, self.num_turbines_full_string), dtype=float)

        self.oss_x = []
        self.oss_y = []

        i = 0
        for oss in self.location_data.substation_id.unique():
            layout = self.location_data[self.location_data.substation_id == oss]
            string_id = np.sort(layout.string.unique())
            string_id += 0 if i == 0 else i

            x = layout.substation_longitude.values[0]
            y = layout.substation_latitude.values[0]
            self.oss_x.append(x)
            self.oss_y.append(y)
            self.location_data_x[string_id, 0] = x
            self.location_data_y[string_id, 0] = y

            for string in string_id:
                data = layout[layout.string == string - i]
                order = data["order"].values
                self.location_data_x[string, order + 1] = data.turbine_longitude.values[order]
                self.location_data_y[string, order + 1] = data.turbine_latitude.values[order]
                self.sections_cable_lengths[string, order] = data.cable_length.values[order]
                self.sections_bury_speeds[string, order] = data.bury_speed.values[order]
            i += string + 1

        # Ensure any point in array without a turbine is set to None
        no_turbines = self.location_data_x == 0
        self.location_data_x[no_turbines] = None
        self.location_data_y[no_turbines] = None

        self.sections_cable_lengths[no_turbines[:, 1:]] = None
        self.sections_bury_speeds[no_turbines[:, 1:]] = None
        self._check_optional_input()

        self.coordinates = np.dstack((self.location_data_x, self.location_data_y))

        # Create the distances between each subsequent turbine in a string
        if self.distance:
            self.sections_distance = self._compute_euclidean_distance()
        else:
            self.sections_distance = self._compute_haversine_distance()

    def run(self):
        self._initialize_cables()
        self.create_strings()
        self._initialize_custom_data()
        self._create_windfarm_layout()
        self._create_cable_section_lengths()

    @property
    def cable_lengths_by_type_speed(self):
        """
        Creates a dictionary of tuples with section lengths and cable burying
        speeds if non-zero entries were provided, otherwise this is equal to
        `cable_lengths_by_type`.

        Returns
        -------
        dict
            A dictionary of the section lengths and burying speeds required for
            each type of cable to fully connect the array cabling system.
            E.g.: {`Cable.name`: [(float, float)]}
        """

        if self.sections_bury_speeds.sum() == 0:
            return self.cable_lengths_by_type
        cables = {
            name: list(
                zip(
                    self.sections_cable_lengths[self.sections_cables == name],
                    self.sections_bury_speeds[self.sections_cables == name],
                )
            )
            for name in self.cables
        }
        return cables



"""Common processes and cargo types for Offshore Substation installations."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from marmot import process

from wisdem.orbit.core import Cargo
from wisdem.orbit.core.logic import stabilize, jackdown_if_required
from wisdem.orbit.core.defaults import process_times as pt
from wisdem.orbit.phases.install.monopile_install.common import (
    bolt_transition_piece,
    cure_transition_piece_grout,
    pump_transition_piece_grout,
)


class Topside(Cargo):
    """Topside Cargo"""

    def __init__(self, mass=None, deck_space=None, **kwargs):
        """
        Creates an instance of `Topside`.
        """

        self.mass = mass
        self.deck_space = deck_space

    @staticmethod
    def fasten(**kwargs):
        """Returns time required to fasten a topside at port."""

        key = "topside_fasten_time"
        time = kwargs.get(key, pt[key])

        return "Fasten Topside", time

    @staticmethod
    def release(**kwargs):
        """Returns time required to release topside from fastenings."""

        key = "topside_release_time"
        time = kwargs.get(key, pt[key])

        return "Release Topside", time


class Jacket(Cargo):
    """Jacket Cargo"""

    pass


@process
def lift_topside(vessel, **kwargs):
    """
    Calculates time required to lift topside at site.

    Parameters
    ----------
    vessel : Vessel
        Vessel to perform action.

    Yields
    ------
    vessel.task representing time to "Lift Topside".
    """

    lift_height = 5  # small lift just to clear the deck
    crane_rate = vessel.crane.crane_rate(**kwargs)
    lift_time = lift_height / crane_rate

    yield vessel.task_wrapper("Lift Topside", lift_time, constraints=vessel.operational_limits)


@process
def attach_topside(vessel, **kwargs):
    """
    Returns time required to attach topside at site.

    Parameters
    ----------
    vessel : Vessel
        Vessel to perform action.
    topside_attach_time : int | float
        Time required to attach topside.

    Yields
    ------
    vessel.task representing time to "Attach Topside".
    """

    _ = vessel.crane

    key = "topside_attach_time"
    attach_time = kwargs.get(key, pt[key])

    yield vessel.task_wrapper("Attach Topside", attach_time, constraints=vessel.operational_limits)


@process
def install_topside(vessel, topside, **kwargs):
    """
    Substation topside installation process.
    Subprocesses:
    - Crane reequip
    - Lift topside
    - Attach topside to substructure
    - Pump grout
    - Cure grout

    Parameters
    ----------
    env : Environment
    vessel : Vessel
    topsdie : dict
    """

    connection = kwargs.get("topside_connection_type", "bolted")
    reequip_time = vessel.crane.reequip(**kwargs)

    yield vessel.task_wrapper(
        "Crane Reequip",
        reequip_time,
        constraints=vessel.transit_limits,
        **kwargs,
    )
    yield lift_topside(vessel)
    yield attach_topside(vessel)

    if connection == "bolted":
        yield bolt_transition_piece(vessel, **kwargs)

    elif connection == "grouted":
        yield pump_transition_piece_grout(vessel, **kwargs)
        yield cure_transition_piece_grout(vessel, **kwargs)

    else:
        raise Exception(
            f"Transition piece connection type '{connection}'" "not recognized. Must be 'bolted' or 'grouted'."
        )

    yield jackdown_if_required(vessel, **kwargs)



import os
import sys
import time

from wisdem import run_wisdem

"""
Example showing how WISDEM values can be changed programmatically in Python.

This uses the `overridden_values` dict given to `run_wisdem`.
Specifically, you can supply a dictionary of values to overwrite after
setup is called.
"""


mydir = os.path.dirname(os.path.realpath(__file__))  # get path to this file
fname_wt_input = os.path.join(os.path.dirname(mydir), "02_reference_turbines", "IEA-15-240-RWT.yaml")
fname_modeling_options = mydir + os.sep + "modeling_options.yaml"
fname_analysis_options = mydir + os.sep + "analysis_options.yaml"

wt_opt, modeling_options, analysis_options = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)
print(f"Tip deflection: {wt_opt['rotorse.rs.tip_pos.tip_deflection'][0]} meters")


# Construct a dict with values to overwrite
overridden_values = {}
overridden_values["rotorse.wt_class.V_mean_overwrite"] = 11.5

# Run the modified simulation with the overwritten values
wt_opt, modeling_options, opt_options = run_wisdem(
    fname_wt_input,
    fname_modeling_options,
    fname_analysis_options,
    overridden_values=overridden_values,
)
print(f"Tip deflection: {wt_opt['rotorse.rs.tip_pos.tip_deflection'][0]} meters")



"""Tests for the Monopile Wisdem API"""

__author__ = ["Jake Nunemaker"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


import openmdao.api as om

from wisdem.orbit.api.wisdem import Orbit


def test_wisdem_monopile_api_default():
    prob = om.Problem(reports=False)
    prob.model = Orbit(floating=False, jacket=False, jacket_legs=0)
    prob.setup()

    prob.run_model()

    prob.model.list_inputs()
    prob.model.list_outputs()


def test_wisdem_jacket_api_default():
    prob = om.Problem(reports=False)
    prob.model = Orbit(floating=False, jacket=True, jacket_legs=3)
    prob.setup()

    prob.run_model()

    prob.model.list_inputs()
    prob.model.list_outputs()


def test_wisdem_floating_api_default():
    prob = om.Problem(reports=False)
    prob.model = Orbit(floating=True, jacket=False, jacket_legs=0)
    prob.setup()

    prob.run_model()

    prob.model.list_inputs()
    prob.model.list_outputs()



# Import the modules
import numpy as np
import openmdao.api as om

import wisdem.commonse.fileIO as fio
from wisdem.drivetrainse.generator import Generator

# ---

# Problem setup

# Whether or not to run optimization
opt_flag = False

# Number of points in efficiency table
n_pc = 20

# Initialize problem instance
prob = om.Problem(reports=False)
prob.model = Generator(design="pmsg_outer", n_pc=n_pc)
# ---

# Pose optimization problem
if opt_flag:
    # Add optimizer and set-up problem (using user defined input on objective function)
    prob.driver = om.ScipyOptimizeDriver()
    prob.driver.options["optimizer"] = "SLSQP"
    prob.driver.options["tol"] = 1e-6

    # prob.driver = om.pyOptSparseDriver()
    # prob.driver.options["optimizer"] = 'CONMIN' # 'SNOPT'

    # Specificiency target efficiency(%)
    Eta_Target = 0.955

    # Design variables
    prob.model.add_design_var("rad_ag", lower=3.0, upper=6.0)
    prob.model.add_design_var("len_s", lower=1.5, upper=2.5)
    prob.model.add_design_var("h_s", lower=0.1, upper=1.00)
    prob.model.add_design_var("p", lower=50.0, upper=100.0, ref=1e2)
    prob.model.add_design_var("h_m", lower=0.01, upper=0.2, ref=1e-2)
    prob.model.add_design_var("h_yr", lower=0.035, upper=0.22, ref=1e-2)
    prob.model.add_design_var("h_ys", lower=0.035, upper=0.22, ref=1e-2)
    prob.model.add_design_var("B_tmax", lower=1, upper=2.0)
    prob.model.add_design_var("t_r", lower=0.05, upper=0.3, ref=1e-2)
    prob.model.add_design_var("t_s", lower=0.05, upper=0.3, ref=1e-2)
    prob.model.add_design_var("h_ss", lower=0.04, upper=0.2, ref=1e-2)
    prob.model.add_design_var("h_sr", lower=0.04, upper=0.2, ref=1e-2)

    # Constraints
    prob.model.add_constraint("B_symax", lower=0.0, upper=2.0)
    prob.model.add_constraint("B_rymax", lower=0.0, upper=2.0)
    prob.model.add_constraint("b_t", lower=0.01, ref=1e-2)
    prob.model.add_constraint("B_g", lower=1.19, upper=1.2)
    prob.model.add_constraint("A_Cuscalc", lower=5.0, upper=600, ref=1e1)
    prob.model.add_constraint("K_rad", lower=0.15, upper=0.3)
    prob.model.add_constraint("Slot_aspect_ratio", lower=4.0, upper=10.0)
    prob.model.add_constraint("generator_efficiency", lower=Eta_Target, indices=[-1])
    prob.model.add_constraint("A_1", upper=110000.0, ref=1e6, indices=[-1])
    prob.model.add_constraint("T_e", lower=21.03e6, upper=21.1e6, ref=1e6)
    prob.model.add_constraint("J_actual", lower=3, upper=6, indices=[-1])
    # prob.model.add_constraint("con_uar", lower=1e-2, ref=1e-2)
    # prob.model.add_constraint("con_yar", lower=1e-2, ref=1e-2)
    # prob.model.add_constraint("con_uas", lower=1e-2, ref=1e-2)
    # prob.model.add_constraint("con_yas", lower=1e-2, ref=1e-2)

    Objective_function = "generator_cost"
    prob.model.add_objective(Objective_function, scaler=1e-5)
    # ----

# Input parameter initialization
prob.setup()

# Specify Target machine parameters
prob["machine_rating"] = 15e6
prob["shaft_rpm"] = np.linspace(5.0, 7.56, n_pc)
prob["rated_torque"] = 21030561.0
prob["P_mech"] = 15354206.45251639

# Drivetrain
prob["D_shaft"] = 3.0
prob["D_nose"] = 2.2

# Generator inputs
prob["B_g"] = 1.38963289
prob["B_r"] = 1.279
prob["B_rymax"] = 1.63478983
prob["B_symax"] = 1.63455514
prob["B_tmax"] = 1.88378905
prob["E_p"] = 3300.0 / np.sqrt(3)
prob["L_s"] = 0.01138752
prob["N_c"] = 2
prob["P_Fe0e"] = 1.0
prob["P_Fe0h"] = 4.0
prob["R_s"] = 0.02457052
prob["S"] = 240
prob["S_N"] = -0.002
prob["alpha_p"] = 1.0995574287564276  # 0.5*np.pi*0.7
prob["b"] = 2.0
prob["b_m"] = 0.1288363023
prob["b_r_tau_r"] = 0.45
prob["b_ro"] = 0.004
prob["b_s"] = 0.05121796888
prob["b_s_tau_s"] = 0.45
prob["b_so"] = 0.004
prob["b_t"] = 0.08159225451
prob["c"] = 5.0
prob["cofi"] = 0.85
prob["freq"] = 60.0
prob["h_0"] = 0.005
prob["h_i"] = 0.004
prob["h_m"] = 0.0995385122
prob["h_s"] = 0.37694491492
prob["h_sr"] = 0.04677
prob["h_ss"] = 0.04651
prob["h_sy0"] = 0.0
prob["h_t"] = 0.3859449149
prob["h_w"] = 0.005
prob["h_yr"] = 0.0361759511
prob["h_ys"] = 0.03618114528
prob["k_fes"] = 0.8
prob["k_fillr"] = 0.55
prob["k_fills"] = 0.65
prob["k_s"] = 0.2
prob["len_s"] = 2.23961662
prob["m"] = 3
prob["mu_0"] = 1.2566370614359173e-06  # np.pi*4e-7
prob["mu_r"] = 1.06
prob["p"] = 100
prob["phi"] = 1.5707963267948966  # 90 deg
prob["q1"] = 5
prob["q2"] = 4
prob["rad_ag"] = 0.5 * 10.25246718
prob["ratio_mw2pp"] = 0.8
prob["resist_Cu"] = 2.52e-8  # 1.8e-8*1.4
prob["sigma"] = 60e3
prob["t_r"] = 0.061
prob["t_s"] = 0.061
prob["tau_p"] = 0.1610453779
prob["tau_s"] = 0.1339360726
prob["u_allow_pcent"] = 8.5
prob["y_allow_pcent"] = 1.0
prob["y_tau_p"] = 0.8  # 12./15.
prob["y_tau_pr"] = 0.8333333  # 10. / 12
prob["z_allow_deg"] = 0.05

# Specific costs
prob["C_Cu"] = 4.786  # Unit cost of Copper $/kg
prob["C_Fe"] = 0.556  # Unit cost of Iron $/kg
prob["C_Fes"] = 0.50139  # specific cost of Structural_mass $/kg
prob["C_PM"] = 95.0

# Material properties
prob["rho_Fe"] = 7700.0  # Steel density Kg/m3
prob["rho_Fes"] = 7850  # structural Steel density Kg/m3
prob["rho_Copper"] = 8900.0  # copper density Kg/m3
prob["rho_PM"] = 7450.0  # typical density Kg/m3 of neodymium magnets (added 2019 09 18) - for pmsg_[disc|arms]
prob["E"] = 2e11
prob["G"] = 79.3e9
# ---

# Run optimization and print results
if opt_flag:
    prob.model.approx_totals()
    prob.run_driver()
else:
    prob.run_model()

prob.model.list_inputs(units=True)
prob.model.list_outputs(units=True)
print("Efficiency table:")
print(" rpm     converter  transformer  generator   ")
print("-------------------------------------------")
print(np.c_[prob["shaft_rpm"],
            prob["converter_efficiency"],
            prob["transformer_efficiency"],
            prob["generator_efficiency"],
            ])
# --
# fio.save_data('PMSG_OUTER', prob, npz_file=False, mat_file=False, xls_file=True)
